import pytest

from tomwer.core.process.reconstruction.nabu.utils import (
    get_recons_volume_identifier,
    get_multi_cor_recons_volume_identifiers,
    nabu_std_err_has_error,
)
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan

_scans = (
    NXtomoScan(
        scan="/my_scan_file.nx",
        entry="entry_test",
    ),
    EDFTomoScan("/my_scan_folder"),
)


@pytest.mark.parametrize("scan", _scans)
@pytest.mark.parametrize("axis", ("YZ", "XZ", "XY"))
def test_get_recons_volume_identifier(scan, axis):
    """
    test get_recons_volume_identifier behavior
    """

    entry = scan.entry if isinstance(scan, NXtomoScan) else "entry"
    # check some exceptions
    with pytest.raises(ValueError):
        get_recons_volume_identifier(
            file_prefix="volume_rec",
            location="/this/is/a/test",
            file_format="toto",
            slice_index=1080,
            scan=scan,
            axis=axis,
        )

    # check hdf5 reconstructions
    id_rec_vols = get_recons_volume_identifier(
        file_prefix="volume_rec",
        location="/this/is/a/test",
        file_format="hdf5",
        slice_index="1080",
        scan=scan,
        axis=axis,
    )
    assert len(id_rec_vols) == 1
    assert (
        id_rec_vols[0].to_str()
        == f"hdf5:volume:/this/is/a/test/volume_rec_plane_{axis}_001080.hdf5?path={entry}/reconstruction"
    )

    # check edf and jp2k slice reconstructions (are expected to have the same behavior)
    id_rec_vols = get_recons_volume_identifier(
        file_prefix="volume_rec",
        location="/this/is/a/test",
        file_format="edf",
        slice_index="1080",
        scan=scan,
        axis=axis,
    )
    assert len(id_rec_vols) == 1
    assert (
        id_rec_vols[0].to_str() == "edf:volume:/this/is/a/test?file_prefix=volume_rec"
    )

    id_rec_vols = get_recons_volume_identifier(
        file_prefix="volume_recslice",
        location="/this/is/a/test",
        file_format="jp2k",
        slice_index="1080",
        scan=scan,
        axis=axis,
    )
    assert len(id_rec_vols) == 1
    assert (
        id_rec_vols[0].to_str()
        == "jp2k:volume:/this/is/a/test?file_prefix=volume_recslice"
    )

    # check tiff slice reconstructions
    id_rec_vols = get_recons_volume_identifier(
        file_prefix="volume_rec",
        location="/this/is/a/test",
        file_format="tiff",
        slice_index="1080",
        scan=scan,
        axis=axis,
    )
    assert len(id_rec_vols) == 1
    assert (
        id_rec_vols[0].to_str() == "tiff:volume:/this/is/a/test?file_prefix=volume_rec"
    )


@pytest.mark.parametrize("scan", _scans)
def test_get_multi_cor_recons_volume_identifier(scan):
    """
    test the get_multi_cor_recons_volume_identifier function
    """
    entry = scan.entry if isinstance(scan, NXtomoScan) else "entry"

    # dummy test with hdf5
    id_rec_vols = get_multi_cor_recons_volume_identifiers(
        scan=scan,
        slice_index="middle",
        location="/this/is/a/test",
        file_prefix="rec",
        cors=(10, 12),
        file_format="hdf5",
    )
    assert isinstance(id_rec_vols, dict)
    assert len(id_rec_vols) == 2
    assert 10 in id_rec_vols.keys()
    assert 12 in id_rec_vols.keys()
    assert (
        id_rec_vols[10].to_str()
        == f"hdf5:volume:/this/is/a/test/rec_10.000_01024.hdf5?path={entry}/reconstruction"
    )

    # dummy test with tiff
    id_rec_vols = get_multi_cor_recons_volume_identifiers(
        scan=scan,
        slice_index="middle",
        location="/this/is/a/test",
        file_prefix="rec",
        cors=(10, 12),
        file_format="tiff",
    )
    assert isinstance(id_rec_vols, dict)
    assert len(id_rec_vols) == 2
    assert 10 in id_rec_vols.keys()
    assert 12 in id_rec_vols.keys()
    assert id_rec_vols[10].to_str() == "tiff:volume:/this/is/a?file_prefix=test"


def test_nabu_std_err_has_error():
    assert nabu_std_err_has_error(None) is False
    assert nabu_std_err_has_error(b"") is False
    assert nabu_std_err_has_error(b"this is an error") is True
    assert (
        nabu_std_err_has_error(
            b"warnings.warn('creating CUBLAS context to get version num"
        )
        is False
    )
    assert (
        nabu_std_err_has_error(
            b"warnings.warn('creating CUBLAS context to get version num\n this is an error"
        )
        is True
    )

    assert (
        nabu_std_err_has_error(
            b"""/nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/skcuda/cublas.py:284: UserWarning: creating CUBLAS context to get version number
            warnings.warn('creating CUBLAS context to get version number')
            /nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/nabu/cuda/kernel.py:49: UserWarning: The CUDA compiler succeeded, but said the following:
            nvcc warning : The 'compute_35', 'compute_37', 'sm_35', and 'sm_37' architectures are deprecated, and may be removed in a future release (Use -Wno-deprecated-gpu-targets to suppress warning).
            self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
            /nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/pycuda/elementwise.py:47: UserWarning: The CUDA compiler succeeded, but said the following:
            nvcc warning : The 'compute_35', 'compute_37', 'sm_35', and 'sm_37' architectures are deprecated, and may be removed in a future release (Use -Wno-deprecated-gpu-targets to suppress warning).
            return SourceModule(
            /nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/nabu/cuda/kernel.py:49: UserWarning: The CUDA compiler succeeded, but said the following:
            nvcc warning : The 'compute_35', 'compute_37', 'sm_35', and 'sm_37' architectures are deprecated, and may be removed in a future release (Use -Wno-deprecated-gpu-targets to suppress warning).
            self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
            /nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/pycuda/elementwise.py:47: UserWarning: The CUDA compiler succeeded, but said the following:
            nvcc warning : The 'compute_35', 'compute_37', 'sm_35', and 'sm_37' architectures are deprecated, and may be removed in a future release (Use -Wno-deprecated-gpu-targets to suppress warning).
            return SourceModule(
            /nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/nabu/cuda/kernel.py:49: UserWarning: The CUDA compiler succeeded, but said the following:
            nvcc warning : The 'compute_35', 'compute_37', 'sm_35', and 'sm_37' architectures are deprecated, and may be removed in a future release (Use -Wno-deprecated-gpu-targets to suppress warning).
            self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
            /nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/nabu/cuda/kernel.py:49: UserWarning: The CUDA compiler succeeded, but said the following:
            nvcc warning : The 'compute_35', 'compute_37', 'sm_35', and 'sm_37' architectures are deprecated, and may be removed in a future release (Use -Wno-deprecated-gpu-targets to suppress warning).
            self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
            /nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/nabu/cuda/kernel.py:49: UserWarning: The CUDA compiler succeeded, but said the following:
            nvcc warning : The 'compute_35', 'compute_37', 'sm_35', and 'sm_37' architectures are deprecated, and may be removed in a future release (Use -Wno-deprecated-gpu-targets to suppress warning).
            kernel.cu(111): warning #1215-D: function "tex2D(texture<T, 2, cudaReadModeElementType>, float, float) [with T=float]"
            /cvmfs/hpc.esrf.fr/software/packages/ubuntu20.04/x86_64/cuda/11.8.0//bin/../targets/x86_64-linux/include/texture_fetch_functions.h(198): here was declared deprecated
            kernel.cu(112): warning #1215-D: function "tex2D(texture<T, 2, cudaReadModeElementType>, float, float) [with T=float]"
            /cvmfs/hpc.esrf.fr/software/packages/ubuntu20.04/x86_64/cuda/11.8.0//bin/../targets/x86_64-linux/include/texture_fetch_functions.h(198): here was declared deprecated
            kernel.cu(113): warning #1215-D: function "tex2D(texture<T, 2, cudaReadModeElementType>, float, float) [with T=float]"
            /cvmfs/hpc.esrf.fr/software/packages/ubuntu20.04/x86_64/cuda/11.8.0//bin/../targets/x86_64-linux/include/texture_fetch_functions.h(198): here was declared deprecated
            kernel.cu(114): warning #1215-D: function "tex2D(texture<T, 2, cudaReadModeElementType>, float, float) [with T=float]"
            /cvmfs/hpc.esrf.fr/software/packages/ubuntu20.04/x86_64/cuda/11.8.0//bin/../targets/x86_64-linux/include/texture_fetch_functions.h(198): here was declared deprecated
            self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
            /nobackup/lbs0511/tomotools/venvs/2023_10_04/lib/python3.8/site-packages/pycuda/elementwise.py:47: UserWarning: The CUDA compiler succeeded, but said the following:
            nvcc warning : The 'compute_35', 'compute_37', 'sm_35', and 'sm_37' architectures are deprecated, and may be removed in a future release (Use -Wno-deprecated-gpu-targets to suppress warning).
            return SourceModule("""
        )
        is False
    )

    assert (
        nabu_std_err_has_error(
            b"""/cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_02_26/lib/python3.11/site-packages/cupyx/jit/_interface.py:173: FutureWarning: cupyx.jit.rawkernel is experimental. The interface can change in the future.
        cupy._util.experimental('cupyx.jit.rawkernel')
        /cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_02_26/lib/python3.11/site-packages/skcuda/cublas.py:284: UserWarning: creating CUBLAS context to get version number
        warnings.warn('creating CUBLAS context to get version number')"""
        )
        is False
    )

    assert (
        nabu_std_err_has_error(
            b"""/cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_04_09/lib/python3.11/site-packages/cupyx/jit/_interface.py:173: FutureWarning: cupyx.jit.rawkernel is experimental. The interface can change in the future.
        cupy._util.experimental('cupyx.jit.rawkernel')
        z_translation Not supported for EDF
        y_translation Not supported for EDF
        x_translation Not supported for EDF
        /cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_04_09/lib/python3.11/site-packages/skcuda/cublas.py:284: UserWarning: creating CUBLAS context to get version number
        warnings.warn('creating CUBLAS context to get version number')
        unable to load metadata from /lbsram/data/id16b/inhouse1/comm_24apr/tomwerUpdate/testTomwer04/reconstructed_volumes/testTomwer04slice_pag_001080_db0500_plane_XY.vol.info"""
        )
        is False
    )

    assert (
        nabu_std_err_has_error(
            b"""/cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_04_19/lib/python3.11/site-packages/cupyx/jit/_interface.py:173: FutureWarning: cupyx.jit.rawkernel is experimental. The interface can change in the future.
        cupy._util.experimental('cupyx.jit.rawkernel')
        /cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_04_19/lib/python3.11/site-packages/nabu/cuda/kernel.py:49: UserWarning: The CUDA compiler succeeded, but said the following:
        kernel.cu(228): warning #177-D: variable "x_stop_other" was declared but never referenced

        kernel.cu(230): warning #177-D: variable "y_stop_other" was declared but never referenced


        self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
        /cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_04_19/lib/python3.11/site-packages/skcuda/cublas.py:284: UserWarning: creating CUBLAS context to get version number
        warnings.warn('creating CUBLAS context to get version number')
        /cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_04_19/lib/python3.11/site-packages/nabu/cuda/kernel.py:49: UserWarning: The CUDA compiler succeeded, but said the following:
        kernel.cu(228): warning #177-D: variable "x_stop_other" was declared but never referenced

        kernel.cu(230): warning #177-D: variable "y_stop_other" was declared but never referenced


        self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
        /cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_04_19/lib/python3.11/site-packages/nabu/cuda/kernel.py:49: UserWarning: The CUDA compiler succeeded, but said the following:
        kernel.cu(131): warning #1215-D: function "tex2D(texture<T, 2, cudaReadModeElementType>, float, float) [with T=float]"
        /cvmfs/hpc.esrf.fr/software/packages/ubuntu20.04/x86_64/cuda/11.7.1/bin/../targets/x86_64-linux/include/texture_fetch_functions.h(198): here was declared deprecated

        kernel.cu(132): warning #1215-D: function "tex2D(texture<T, 2, cudaReadModeElementType>, float, float) [with T=float]"
        /cvmfs/hpc.esrf.fr/software/packages/ubuntu20.04/x86_64/cuda/11.7.1/bin/../targets/x86_64-linux/include/texture_fetch_functions.h(198): here was declared deprecated

        kernel.cu(133): warning #1215-D: function "tex2D(texture<T, 2, cudaReadModeElementType>, float, float) [with T=float]"
        /cvmfs/hpc.esrf.fr/software/packages/ubuntu20.04/x86_64/cuda/11.7.1/bin/../targets/x86_64-linux/include/texture_fetch_functions.h(198): here was declared deprecated

        kernel.cu(134): warning #1215-D: function "tex2D(texture<T, 2, cudaReadModeElementType>, float, float) [with T=float]"
        /cvmfs/hpc.esrf.fr/software/packages/ubuntu20.04/x86_64/cuda/11.7.1/bin/../targets/x86_64-linux/include/texture_fetch_functions.h(198): here was declared deprecated

        kernel.cu(21): warning #177-D: function "linear_interpolation" was declared but never referenced


        self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
        /cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_04_19/lib/python3.11/site-packages/cupy/cuda/compiler.py:233: PerformanceWarning: Jitify is performing a one-time only warm-up to populate the persistent cache, this may take a few seconds and will be improved in a future release...
        jitify._init_module()"""
        )
        is False
    )

    assert (
        nabu_std_err_has_error(
            b"""pytools.persistent_dict: unable to import 'siphash24.siphash13', """
        )
        is False
    )

    assert (
        nabu_std_err_has_error(
            b"""CompilerWarning: Non-empty compiler output encountered. Set the environment variable PYOPENCL_COMPILER_OUTPUT=1 to see more.
        _create_built_program_from_source_cached(
        /cvmfs/tomo.esrf.fr/software/packages/linux/x86_64/tomotools/2024_09_06/lib/python3.11/site-packages/pyopencl/cache.py:499: CompilerWarning: Non-empty compiler output encountered. Set the environment variable PYOPENCL_COMPILER_OUTPUT=1 to see more.
        prg.build(options_bytes, devices)"""
        )
        is False
    )



# coding: utf-8
from __future__ import annotations


import logging

from tomwer.gui.control.datalist import VolumeList
from tomwer.gui.dialog.QVolumeDialog import QVolumeDialog

from .selectorwidgetbase import _SelectorWidget

logger = logging.getLogger(__name__)


class VolumeSelectorWidget(_SelectorWidget):
    """Widget used to select a volume on a list"""

    def _buildDataList(self):
        return VolumeList(self)

    def _callbackAddData(self):
        dialog = QVolumeDialog(self)

        if not dialog.exec_():
            dialog.close()
            return

        volume = dialog.getVolume()
        if volume is not None:
            added_objs = self.add(volume)
            self.setMySelection(added_objs)
        self.sigUpdated.emit()



"""Utils to load metadata (from a volume at the moment)
"""

import logging
from silx.gui import qt
from tomwer.core.volume.volumefactory import VolumeFactory

_logger = logging.getLogger(__name__)

__all__ = [
    "VolumeMetadataLoader",
]


class VolumeMetadataLoader(qt.QThread):
    """Thread to load metadata from a given volume"""

    def __init__(self, parent, tomo_obj_identifier: str):
        super().__init__(parent=parent)
        self.volume = VolumeFactory.create_tomo_object_from_identifier(
            tomo_obj_identifier
        )
        self.metadata = None

    def run(self):
        try:
            self.metadata = self.volume.load_metadata()
        except IOError:
            self.metadata = None
        except Exception as e:
            _logger.error(
                "Fail to load metadata from %s. Error is %s",
                self.volume.get_identifier(),
                e,
            )
            self.metadata = None



# coding: utf-8

"""
This module is used to manage the rsync between files for transfert.
"""

from __future__ import annotations


import logging
import time

from rsyncmanager import RSyncManager
from silx.gui import qt

logger = logging.getLogger(__name__)


class RSyncWorker(qt.QThread):
    """Thread which call for a synchronization on a folder each n seconds"""

    def __init__(self, src_dir, dst_dir, delta_time):
        qt.QThread.__init__(self)
        self._src_dir = src_dir
        self._dst_dir = dst_dir
        self._delta_tile = delta_time
        self._stop = False

    def stop(self):
        self._stop = True

    def _process_sync(self):
        RSyncManager().sync_file(
            source=self._src_dir, target=self._dst_dir, delete=False
        )
        time.sleep(self._delta_tile)

    def run(self):
        while not self._stop:
            self._process_sync()


class BlissSequenceRSyncWorker(RSyncWorker):
    """Thread to synchronize a bliss sequence"""

    def __init__(
        self,
        src_dir,
        dst_dir,
        src_proposal_file,
        dst_proposal_file,
        src_sample_file,
        dst_sample_file,
        delta_time,
    ):
        super(BlissSequenceRSyncWorker, self).__init__(
            src_dir=src_dir, dst_dir=dst_dir, delta_time=delta_time
        )
        self._bliss_master_file = src_proposal_file
        self._dst_bliss_master_file = dst_proposal_file
        self._bliss_sample_file = src_sample_file
        self._dst_bliss_sample_file = dst_sample_file

    def _process_sync(self):
        super(BlissSequenceRSyncWorker, self)._process_sync()
        if self._bliss_master_file is not None:
            RSyncManager().sync_file(
                source=self._bliss_master_file, target=self._dst_bliss_master_file
            )
        if self._bliss_sample_file is not None:
            RSyncManager().sync_file(
                source=self._bliss_sample_file, target=self._dst_bliss_sample_file
            )


class BlissSequenceRSyncWorkerFinalizer(qt.QThread):
    def __init__(
        self,
        source_scans,
        target_scans,
        proposal_file,
        target_proposal_file,
        sample_file,
        target_sample_file,
        nx_file,
        target_nx_file,
        block,
        callback,
        callback_parameters,
        parallel,
        delta_time=0.5,
    ):
        self._sources_scans = source_scans
        self._target_scans = target_scans
        self._proposal_file = proposal_file
        self._target_proposal_file = target_proposal_file
        self._sample_file = sample_file
        self._target_sample_file = target_sample_file
        self._nx_file = nx_file
        self._target_nx_file = target_nx_file
        self._block = block
        self._callback = callback
        self._callback_parameters = callback_parameters
        self._parparallel = parallel

        self._n_callback_scans = 0
        self._n_callback_proposal = 0
        self._n_callback_sample = 0
        self._delta_time = delta_time

    def scan_transfert_completed(self):
        return self._n_callback_scans == len(self._sources_scans)

    def proposal_file_completed(self):
        if self._proposal_file is None:
            return True
        else:
            return self._n_callback_proposal == 1

    def sample_file_completed(self):
        if self._sample_file is None:
            return True
        else:
            return self._n_callback_sample == 1

    def transfert_completed(self):
        return (
            self.proposal_file_completed()
            and self.sample_file_completed()
            and self.scan_transfert_completed()
        )

    def _callback_scans(self, *args, **kwargs):
        self._n_callback_scans = self._n_callback_scans + 1

    def _callaback_sample_file(self, *args, **kwargs):
        self._n_callback_sample = 1

    def _callback_proposal_file(self, *args, **kwargs):
        self._n_callback_proposal = 1

    def run(self):
        # no callback
        if self._sample_file is not None:
            RSyncManager().sync_file(
                source=self._sample_file, target=self._target_sample_file
            )

        if self._proposal_file is not None:
            RSyncManager().sync_file(
                source=self._proposal_file, target=self._target_proposal_file
            )

        for scan_in, scan_out in zip(self._sources_scans, self._target_scans):
            RSyncManager().sync_file(
                source=scan_in,
                target=scan_out,
                callback=self._callback_scans,
                delete=True,
            )

        while not self.transfert_completed():
            time.sleep(self._delta_time)

        if self._callback:
            if self._callback_parameters is not None:
                self._callback(*self._callback_parameters)
            else:
                self._callback()



"""
This module provides global definitions and functions to manage dark and flat fields
especially for tomo experiments and workflows
"""

from __future__ import annotations

import logging
import os
from queue import Queue

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.utils.deprecation import deprecated_warning

from tomoscan.framereducer.target import REDUCER_TARGET
from tomoscan.framereducer.method import ReduceMethod

import tomwer.version
from tomwer.core import settings
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.process.task import Task
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.utils import docstring

from . import params as dkrf_reconsparams

logger = logging.getLogger(__name__)


class DarkRefsTask(
    Task,
    SuperviseProcess,
    Queue,  # TODO: fixme: this does not make much sense for a task
    input_names=("data",),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """Compute median/mean dark and ref from originals (dark and ref files)"""

    WHAT_REF = "refs"
    WHAT_DARK = "dark"

    VALID_WHAT = (WHAT_REF, WHAT_DARK)
    """Tuple of valid option for What"""

    info_suffix = ".info"

    TOMO_N = "TOMO_N"

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        """

        :param file_ext:
        :param reconsparams: reconstruction parameters
        """
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        SuperviseProcess.__init__(self, process_id=process_id)
        Queue.__init__(self)

        if inputs is None:
            inputs = {}
        if "recons_params" in inputs or "reconsparams" in inputs:
            raise ValueError(
                "wrong key: recons_params / reconsparams use dark_ref_params instead"
            )
        self._recons_params = inputs.get("dark_ref_params", None)
        if self._recons_params is None:
            self._recons_params = dkrf_reconsparams.DKRFRP()
        elif isinstance(self._recons_params, dict):
            self._recons_params = dkrf_reconsparams.DKRFRP.from_dict(
                self._recons_params
            )
        if not isinstance(self._recons_params, dkrf_reconsparams.DKRFRP):
            raise TypeError(
                f"'reconsparams' should be an instance of {dkrf_reconsparams.DKRFRP} or {dict}. Not {type(self._recons_params)}"
            )
        self._file_ext = inputs.get("file_ext", ".edf")
        if not type(self._file_ext) is str:
            raise TypeError("'file_ext' is expected to be a string")

        self._forceSync = inputs.get("force_sync", False)
        self.__new_hdf5_entry_created = False
        "used to know if the process has generated a new entry or not"

    @property
    def recons_params(self):
        return self._recons_params

    def set_recons_params(self, recons_params):
        if isinstance(recons_params, dkrf_reconsparams.DKRFRP):
            self._recons_params = recons_params
        else:
            raise TypeError(
                "recons_params should be an instance of " "ReconsParams or DKRFRP"
            )

    def setPatternRecons(self, pattern):
        self._patternReconsFile = pattern

    def setForceSync(self, b):
        self._forceSync = True

    @staticmethod
    def getRefHSTFiles(directory, prefix, file_ext=".edf"):
        """

        :return: the list of existing refs files in the directory according to
                 the file pattern.
        """
        assert isinstance(directory, str)
        res = []
        if os.path.isdir(directory) is False:
            logger.error(
                directory + " is not a directory. Cannot extract " "RefHST files"
            )
            return res

        for file in os.listdir(directory):
            if file.startswith(prefix) and file.endswith(file_ext):
                res.append(os.path.join(directory, file))
                assert os.path.isfile(res[-1])
        return res

    @staticmethod
    def getDarkHSTFiles(directory, prefix, file_ext=".edf"):
        """

        :return: the list of existing refs files in the directory according to
                 the file pattern.
        """
        res = []
        if os.path.isdir(directory) is False:
            logger.error(
                directory + " is not a directory. Cannot extract " "DarkHST files"
            )
            return res
        for file in os.listdir(directory):
            _prefix = prefix
            if prefix.endswith(file_ext):
                _prefix = prefix.rstrip(file_ext)
            if file.startswith(_prefix) and file.endswith(file_ext):
                _file = file.lstrip(_prefix).rstrip(file_ext)
                if _file == "" or _file.isnumeric() is True:
                    res.append(os.path.join(directory, file))
                    assert os.path.isfile(res[-1])
        return res

    @staticmethod
    def getDarkPatternTooltip():
        return (
            "define the pattern to find, using the python `re` library.\n"
            "For example: \n"
            "   - `.*conti_dark.*` to filter files containing `conti_dark` sentence\n"
            "   - `darkend[0-9]{3,4}` to filter files named `darkend` followed by three or four digit characters (and having the .edf extension)"
        )

    @staticmethod
    def getRefPatternTooltip():
        return (
            "define the pattern to find, using the python `re` library.\n"
            "For example: \n"
            "   - `.*conti_ref.*` for files containing `conti_dark` sentence\n"
            "   - `ref*.*[0-9]{3,4}_[0-9]{3,4}` to filter files named `ref` followed by any character and ending by X_Y where X and Y are groups of three or four digit characters."
        )

    @staticmethod
    def properties_help():
        return """
        - refs: 'None', 'Median', 'Average', 'First', 'Last' \n
        - dark: 'None', 'Median', 'Average', 'First', 'Last' \n
        """

    def set_configuration(self, properties):
        # No properties stored for now
        if "dark" in properties:
            self._recons_params.dark_calc_method = properties["dark"]
        if "refs" in properties:
            self._recons_params.flat_calc_method = properties["refs"]
        if "_rpSetting" in properties:
            self._recons_params.load_from_dict(properties["_rpSetting"])
        else:
            self._recons_params.load_from_dict(properties)

    @staticmethod
    def get_darks_frm_process_file(
        process_file, entry=None, as_url: bool = False
    ) -> dict | None:
        """

        :param process_file: path to the process file
        :param entry: entry to read in the process file if more than one
        :param as_url: if true then an url will be used instead of a
                            numpy.array
        :return: dictionary with index in the sequence as key and numpy array
                 as value (or url if as_url set to True)
        """
        deprecated_warning(
            name="tomwer.core.process.reconstruction.darkref.DarkRefsTask.get_darks_frm_process_file",
            type_="function",
            reason="removed. Reduced dark / flat are expected to be stored on didicated files and not in tomwer_processes anymore",
            since_version="1.2",
        )
        return None

    @staticmethod
    def get_flats_frm_process_file(
        process_file, entry=None, as_url: bool = False
    ) -> dict | None:
        """

        :param process_file:
        :param entry: entry to read in the process file if more than one
        :param as_url: if true then an url will be used instead of a
                            numpy.array
        :return:
        """
        deprecated_warning(
            name="tomwer.core.process.reconstruction.darkref.DarkRefsTask.get_flats_frm_process_file",
            type_="function",
            reason="removed. Reduced dark / flat are expected to be stored on didicated files and not in tomwer_processes anymore",
            since_version="1.2",
        )
        return None

    def run(self):
        if isinstance(self.inputs.data, str):
            try:
                scan = data_identifier_to_scan(self.inputs.data)
            except ValueError:
                # in the case fails to cast str to identifier (data can be a simple folder in the case of dark ref)
                scan = self.inputs.data
        else:
            scan = self.inputs.data

        if scan is None:
            self.outputs.data = None
            return

        if type(scan) is str:
            assert os.path.exists(scan)
            scan = ScanFactory.create_scan_object(scan_path=scan)
        elif isinstance(scan, TomwerScanBase):
            pass
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise TypeError(
                "scan should be an instance of TomoBase or path to " "scan dircetory"
            )
        assert isinstance(self._recons_params, dkrf_reconsparams.DKRFRP)
        assert self._recons_params is not None

        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=DatasetState.ON_GOING,
        )
        logger.processStarted(f"start dark and ref for {scan}")
        if (
            settings.isOnLbsram(scan)
            and is_low_on_memory(settings.get_lbsram_path()) is True
        ):
            mess = (
                "low memory, do compute dark and flat field mean/median "
                "for %s" % scan.path
            )
            logger.processSkipped(mess)
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self,
                state=DatasetState.SKIPPED,
                details=mess,
            )
            self.outputs.data = None
            return

        if not (scan and os.path.exists(scan.path)):
            mess = f"folder {scan.folder} is not existing"
            logger.warning(mess)
            ProcessManager().notify_dataset_state(
                dataset=scan, process=self, state=DatasetState.FAILED, details=mess
            )

            self.outputs.data = None
            return
        whats = (DarkRefsTask.WHAT_REF, DarkRefsTask.WHAT_DARK)
        overwrites = (
            self.recons_params.overwrite_flat,
            self.recons_params.overwrite_dark,
        )
        modes = (
            self.recons_params.flat_calc_method,
            self.recons_params.dark_calc_method,
        )
        has_reduced = (
            scan.reduced_flats not in (None, {}),
            scan.reduced_darks not in (None, {}),
        )

        for what, mode, overwrite, exists in zip(whats, modes, overwrites, has_reduced):
            # if reduced already exists and user didn't asked for overwritting it
            if exists and not overwrite:
                continue
            logger.debug(f"compute {what} using mode {mode} for {scan}")
            try:
                self.compute(scan=scan, target=what, method=mode, overwrite=True)
            except Exception as e:
                info = f"Fail computing dark and flat for {scan}. Reason is {e}"
                self.notify_to_state_to_managed(
                    dataset=scan, state=DatasetState.FAILED, details=info
                )
                logger.processFailed(info)
                self.outputs.data = None
                return
        results = {}
        interpretations = {}
        if (
            self.recons_params.dark_calc_method
            is not dkrf_reconsparams.ReduceMethod.NONE
            and scan.reduced_darks is not None
        ):
            # cast darks and flats keys from int (index) to str
            o_darks = scan.reduced_darks
            o_darks_infos = scan.reduced_darks_infos
            f_darks = {}
            for index, data in o_darks.items():
                f_darks[str(index)] = data
                interpretations["/".join(("darks", str(index)))] = "image"
            results["darks"] = f_darks

            scan.save_reduced_darks(f_darks, darks_infos=o_darks_infos, overwrite=True)
        if (
            self.recons_params.flat_calc_method
            is not dkrf_reconsparams.ReduceMethod.NONE
            and scan.reduced_flats is not None
        ):
            results["flats"] = scan.reduced_flats
            o_flats = scan.reduced_flats
            o_flats_infos = scan.reduced_flats_infos
            f_flats = {}
            for index, data in o_flats.items():
                f_flats[str(index)] = data
                interpretations["/".join(("flats", str(index)))] = "image"
            results["flats"] = f_flats

            scan.save_reduced_flats(f_flats, flats_infos=o_flats_infos, overwrite=True)

        if len(results) > 0:
            # if some processing to be registered
            if scan.process_file is not None and not (
                isinstance(scan, NXtomoScan) and not self.__new_hdf5_entry_created
            ):
                entry = "entry"
                if hasattr(scan, "entry"):
                    entry = scan.entry
                self.register_process(
                    process_file=scan.process_file,
                    entry=entry,
                    configuration=self.recons_params.to_dict(),
                    results=results,
                    interpretations=interpretations,
                    process_index=scan.pop_process_index(),
                    overwrite=True,
                )
        logger.processSucceed(f"Dark and flat reduction succeeded for {scan}")
        self.notify_to_state_to_managed(
            dataset=scan, state=DatasetState.SUCCEED, details=None
        )

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan

    @staticmethod
    def _target_to_reducer_target(target):
        """
        util to insure connection between 'historical' tomwer dark / ref and latest tomoscan FrameReducer
        """
        if target == "refs":
            return REDUCER_TARGET.FLATS
        elif target == "dark":
            return REDUCER_TARGET.DARKS
        return REDUCER_TARGET.from_value(target)

    def compute(
        self, scan, target: REDUCER_TARGET, method: ReduceMethod, overwrite: bool
    ):
        target = self._target_to_reducer_target(target)
        method = ReduceMethod.from_value(method)
        if target is REDUCER_TARGET.DARKS:
            reduced_darks, metadata = scan.compute_reduced_darks(
                reduced_method=method,
                overwrite=overwrite,
                return_info=True,
            )
            scan.set_reduced_darks(darks=reduced_darks, darks_infos=metadata)

            try:
                scan.save_reduced_darks(
                    darks=reduced_darks, darks_infos=metadata, overwrite=True
                )
            except Exception as e:
                logger.error(f"Fail to save reduced darks. Error is {e}")
        elif target is REDUCER_TARGET.FLATS:
            reduced_flats, metadata = scan.compute_reduced_flats(
                reduced_method=method,
                overwrite=overwrite,
                return_info=True,
            )
            scan.set_reduced_flats(flats=reduced_flats, flats_infos=metadata)

            try:
                scan.save_reduced_flats(
                    flats=reduced_flats, flats_infos=metadata, overwrite=True
                )
            except Exception as e:
                logger.error(f"Fail to save reduced flats. Error is {e}")
        else:
            raise RuntimeError(f"{target} not handled")

    @docstring(Task.program_name)
    @staticmethod
    def program_name():
        return "tomwer_dark_refs"

    @docstring(Task.program_version)
    @staticmethod
    def program_version():
        return tomwer.version.version

    @docstring(Task.definition)
    @staticmethod
    def definition():
        return "Compute mean or median dark and refs per each series"


def requires_reduced_dark_and_flat(
    scan: TomwerScanBase, logger_: logging.Logger | None = None
) -> tuple:
    r"""helper function: If no dark / flat are computed yet then will pick the first
    dark and the first flat.

    Expected usage: for tomwer application which requires some time and flat and to avoid some warnings
    within standalones.

    :params scan: scan for which we want to get quick dark and flat
    :params logger\_: if provided will add some warning when attempt to get reduced flat / dark.
    :returns: tuple of what was missing and has been computed
    """
    computed = []
    if scan.reduced_flats in (None, {}):
        # set the first flat found
        recons_params = dkrf_reconsparams.DKRFRP()
        recons_params.overwrite_dark = False
        recons_params.overwrite_flat = False
        recons_params.dark_calc_method = dkrf_reconsparams.ReduceMethod.NONE
        recons_params.flat_calc_method = dkrf_reconsparams.ReduceMethod.FIRST

        drp = DarkRefsTask(
            inputs={
                "data": scan,
                "dark_ref_params": recons_params,
                "serialize_output_data": False,
            }
        )
        if logger_ is not None:
            logger.warning(
                "No 'reduced' flat found. Will try to pick the first flat found as the `calculated` flat."
            )
        drp.run()
        computed.append("flat")

    if scan.reduced_darks in (None, {}):
        # set the first dark found
        recons_params = dkrf_reconsparams.DKRFRP()
        recons_params.overwrite_dark = False
        recons_params.overwrite_flat = False
        recons_params.dark_calc_method = dkrf_reconsparams.ReduceMethod.FIRST
        recons_params.flat_calc_method = dkrf_reconsparams.ReduceMethod.NONE

        drp = DarkRefsTask(
            inputs={
                "data": scan,
                "dark_ref_params": recons_params,
                "serialize_output_data": False,
            }
        )
        if logger_ is not None:
            logger.warning(
                "No 'reduced' dark found. Will try to pick the first flat found as the `calculated` dark."
            )
        drp.run()
        computed.append("dark")

    return tuple(computed)


class DarkRefs(DarkRefsTask):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        deprecated_warning(
            name="tomwer.core.process.reconstruction.darkref.darkref.DarkRefs",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="DarkRefsTask",
        )
        super().__init__(process_id, varinfo, inputs, node_id, node_attrs, execinfo)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Application to display the difference between two frames.
"""

import argparse
import logging
import signal
import sys
import silx
from silx.gui import qt

from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.visualization.diffviewer import DiffFrameViewer

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer nabu [scan_path]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.closeAllWindows()  # needed because get a waiting thread behind
    qt.QApplication.quit()


def main(argv):
    import os

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path", help="Data file to show (h5 file, edf files, spec files)"
    )
    parser.add_argument(
        "--entry", default=None, help="An entry should be specify for hdf5 files"
    )
    parser.add_argument("--second-scan", default=None, help="Second data files")
    parser.add_argument("--second-entry", default=None, help="Second entry")
    parser.add_argument(
        "--use-opengl-plot",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )
    options = parser.parse_args(argv[1:])
    options.scan_path = options.scan_path.rstrip(os.path.sep)

    if options.second_entry is not None and options.second_scan is None:
        options.second_scan = options.scan_path

    increase_max_number_file()

    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()

    # create scan
    scan_1 = ScanFactory.create_scan_object(options.scan_path, entry=options.entry)
    if scan_1 is None:
        raise ValueError(
            "Given scan path is not recognized as a path" "containing a scan"
        )
    if options.second_scan is not None:
        scan_2 = ScanFactory.create_scan_object(
            options.second_scan, entry=options.second_entry
        )
    else:
        scan_2 = None

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    # handle gui
    widget = DiffFrameViewer(parent=None)
    widget.addScan(scan_1)
    if scan_2 is not None:
        widget.addScan(scan_2)
        widget.setRightScan(scan_2)
    # for the application we run for the reconstruction to be finished
    # to give back hand to the user
    widget.setWindowTitle("Frame diff")
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
from __future__ import annotations

import gc
import tempfile

import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.visualization.SampleMovedOW import SampleMovedOW
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSampleMoved(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        super().setUp()
        self._window = SampleMovedOW()
        self._tmp_path = tempfile.mkdtemp()

        self._scan = MockNXtomo(
            scan_path=self._tmp_path,
            create_ini_dark=True,
            create_ini_flat=True,
            create_final_flat=False,
            n_proj=10,
            n_ini_proj=10,
            dim=12,
        ).scan

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        gc.collect()

    def test(self):
        self._window.show()
        self.qWaitForWindowExposed(self._window)

        self._window.updateScan(self._scan)
        self.qapp.processEvents()



# coding: utf-8

from ewokscore.task import Task as EwoksTask


class _SampleMovedPlaceHolder(EwoksTask, input_names=("data",)):
    """
    Task to check if a sample has moved along time / projections
    """

    def run(self):
        pass



from __future__ import annotations

import functools
import logging
import os
import re
from collections import OrderedDict
from glob import glob
from math import ceil
import pathlib
from datetime import datetime

import fabio
import numpy
from tqdm import tqdm
from processview.core.dataset import DatasetIdentifier
from silx.io.url import DataUrl

from tomoscan.esrf.identifier.edfidentifier import (
    EDFTomoScanIdentifier as _EDFTomoScanIdentifier,
)
from tomoscan.esrf.scan.edfscan import EDFTomoScan as _tsEDFTomoScan
from tomoscan.esrf.scan.utils import get_data
from nxtomo.nxobject.nxdetector import FOV

from tomwer.core.process.reconstruction.darkref.settings import (
    DARKHST_PREFIX,
    REFHST_PREFIX,
)
from tomwer.utils import docstring

from .scanbase import TomwerScanBase

_logger = logging.getLogger(__name__)


global counter_rand
counter_rand = 1  # used to be sure to return a unique index on recons slices


class EDFTomoScanIdentifier(_EDFTomoScanIdentifier, DatasetIdentifier):
    def __init__(self, object, folder, file_prefix, metadata=None):
        super().__init__(object, folder, file_prefix)
        DatasetIdentifier.__init__(self, EDFTomoScan.from_identifier, metadata=metadata)

    @staticmethod
    def from_str(identifier):
        from tomoscan.esrf.scan.edfscan import EDFTomoScan

        return _EDFTomoScanIdentifier._from_str_to_single_frame_identifier(
            identifier=identifier,
            SingleFrameIdentifierClass=EDFTomoScanIdentifier,
            ObjClass=EDFTomoScan,
        )

    def long_description(self) -> str:
        """used for processview header tooltip for now"""
        return self.to_str()

    def short_description(self) -> str:
        return f"scan: {os.path.basename(os.path.basename(self.folder))} (EDF)"


class EDFTomoScan(_tsEDFTomoScan, TomwerScanBase):
    """
    Class used to represent a tomography acquisition with hdf5 files.

    :param scan: path of the scan
    """

    def __init__(self, scan, overwrite_proc_file=False, update=True):
        _tsEDFTomoScan.__init__(self, scan=scan)
        TomwerScanBase.__init__(self)
        # register at least the 'default' working directory as a possible reconstruction path
        self.add_reconstruction_path(self.path)

        self._dark = None
        self._process_file = self.get_process_file_name(self)
        self._init_index_process_file(overwrite_proc_file=overwrite_proc_file)

        if scan is not None and update:
            try:
                self.update()
            except IOError:
                # fabio can raise some empty file error when data is not on disk (can be the case for the datawatcher when acquisiton is on going)
                pass
        try:
            reduced_darks, metadata = self.load_reduced_darks(return_info=True)
        except (KeyError, OSError, ValueError):
            # file or key does not exists
            pass
        else:
            self.set_reduced_darks(reduced_darks, darks_infos=metadata)

        try:
            reduced_flats, metadata = self.load_reduced_flats(return_info=True)
        except (KeyError, OSError, ValueError):
            pass
        else:
            self.set_reduced_flats(reduced_flats, flats_infos=metadata)

    @staticmethod
    def is_tomoscan_dir(
        directory: str,
        dataset_basename: str | None = None,
        src_pattern=None,
        dest_pattern=None,
        **kwargs,
    ) -> bool:
        info_file = EDFTomoScan.get_info_file(
            directory=directory, dataset_basename=dataset_basename, kwargs=kwargs
        )
        if src_pattern is not None and dest_pattern is not None:
            infofilenice = info_file.replace(src_pattern, dest_pattern, 1)
            return os.path.isfile(infofilenice) or os.path.isfile(info_file)
        else:
            return os.path.isfile(info_file)

    @property
    def working_directory(self):
        return self.path

    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, EDFTomoScanIdentifier):
            raise TypeError(
                f"identifier should be an instance of {EDFTomoScanIdentifier} not {type(identifier)}"
            )
        return EDFTomoScan(scan=identifier.folder)

    @staticmethod
    def get_process_file_name(scan):
        if scan.path is not None:
            basename = os.path.basename(scan.path)
            basename = "_".join((basename, "tomwer_processes.h5"))
            return os.path.join(scan.path, basename)
        else:
            return None

    def clear_caches(self):
        _tsEDFTomoScan.clear_caches(self)
        TomwerScanBase.clear_caches(self)

    @staticmethod
    def directory_contains_scan(directory, src_pattern, dest_pattern):
        """
        Check if the given directory is holding an acquisition

        :param directory: directory we want to check
        :param src_pattern: buffer name pattern ('lbsram')
        :param dest_pattern: output pattern (''). Needed because some
                             acquisition can split the file produce between
                             two directories. This is the case for edf,
                             where .info file are generated in /data/dir
                             instead of /lbsram/data/dir
        :return: does the given directory contains any acquisition
        """
        aux = directory.split(os.path.sep)
        info_name = os.path.join(directory, aux[-1] + EDFTomoScan.INFO_EXT)

        if src_pattern:
            info_name = info_name.replace(src_pattern, dest_pattern, 1)
        return os.path.isfile(info_name)

    def data_flat_field_correction(self, data, index=None):
        """
        Apply the flat field correction on the given data.

        :param data: radio to correct
        :return numpy.ndarray: corrected data
        """
        dark = self.getDark()
        flat = self.getFlat()
        return self._flat_field_correction(
            data=data,
            dark=dark,
            flat1=flat,
            flat2=None,
            index_flat1=-1,
            index_flat2=-1,
            index_proj=index,
        )

    def update(self):
        """update list of radio and reconstruction by parsing the scan folder"""
        _tsEDFTomoScan.update(self)
        self.reconstructions = EDFTomoScan.get_reconstructions_paths(self.path)

    def projections_with_angle(self):
        return self.get_proj_angle_url()

    @staticmethod
    def get_reconstructions_paths(scanID, withIndex=False):
        """
        Return the dict of files:
        * fitting with a reconstruction pattern and ending by .edf
        * .vol files

        :param scanID: is the path to the folder of acquisition
        :param withIndex: if False then return a list of slices otherwise
            return a dict with the index of the slice reconstructed.
        """

        def containsDigits(input_):
            return any(char.isdigit() for char in input_)

        if (scanID is None) or (not os.path.isdir(scanID)):
            if withIndex is True:
                return {}
            else:
                return []

        pyhst_files = TomwerScanBase.get_pyhst_recons_file(scanID)
        if pyhst_files is not None:
            return TomwerScanBase.getReconstructedFilesFromParFile(
                pyhst_files, with_index=withIndex
            )
        else:
            folderBasename = os.path.basename(scanID)
            files = {} if withIndex is True else []
            if os.path.isdir(scanID):
                for f in os.listdir(scanID):
                    if (
                        f.endswith(TomwerScanBase.VALID_RECONS_EXTENSION)
                        and f.startswith(folderBasename)
                        and "slice_" in f
                    ):
                        local_str = f
                        for extension in TomwerScanBase.VALID_RECONS_EXTENSION:
                            if local_str.endswith(extension):
                                local_str = local_str.rsplit(extension, 1)[0]
                        if "slice_" in local_str:
                            if "slice_pag_" in local_str:
                                indexStr = local_str.split("slice_pag_")[-1].split("_")[
                                    0
                                ]
                            elif "slice_ctf_" in local_str:
                                indexStr = local_str.split("slice_ctf_")[-1].split("_")[
                                    0
                                ]
                            else:
                                indexStr = local_str.split("slice_")[-1].split("_")[0]
                            if containsDigits(indexStr):
                                gfile = os.path.join(scanID, f)
                                assert os.path.isfile(gfile)
                                if withIndex is True:
                                    files[
                                        EDFTomoScan.get_index_reconstructed(f, scanID)
                                    ] = gfile
                                else:
                                    files.append(gfile)
                    if f.endswith(".vol"):
                        if withIndex is True:
                            files[EDFTomoScan.get_index_reconstructed(f, scanID)] = (
                                os.path.join(scanID, f)
                            )
                        else:
                            files.append(os.path.join(scanID, f))
            return files

    def load_from_dict(self, desc):
        _tsEDFTomoScan.load_from_dict(self, desc)
        TomwerScanBase.load_from_dict(self, desc)
        return self

    def _get_scheme(self):
        """

        :return: scheme to read url
        """
        return "fabio"

    @docstring(TomwerScanBase.get_sinogram)
    @functools.lru_cache(maxsize=16, typed=True)
    def get_sinogram(self, line, subsampling=1, norm_method=None, **kwargs):
        """

        extract the sinogram from projections

        :param line: which sinogram we want
        :param subsampling: subsampling to apply if any. Allows to skip some io
        :return: sinogram from the radio lines
        """
        _logger.info(
            f"compute sinogram for line {line} of {self.path} (subsampling: {subsampling})"
        )
        assert isinstance(line, int)
        if self.tomo_n is not None and self.dim_2 is not None and line > self.dim_2:
            raise ValueError("requested line %s is not in the scan")
        else:
            y_dim = ceil(self.tomo_n / subsampling)
            sinogram = numpy.empty((y_dim, self.dim_1))
            proj_urls = self.get_proj_angle_url()
            assert len(proj_urls) >= self.tomo_n
            proj_sort = list(proj_urls.keys())
            proj_sort = list(filter(lambda x: not isinstance(x, str), proj_sort))
            proj_sort.sort()
            advancement = tqdm(
                desc=f"compute sinogram for {os.path.basename(self.path)}, line={line}, sampling={subsampling}"
            )
            advancement.total = len(proj_sort)
            for i_proj, proj in enumerate(proj_sort):
                if i_proj % subsampling == 0:
                    url = proj_urls[proj]
                    radio = get_data(url)
                    radio = self.data_flat_field_correction(radio)
                    sinogram[i_proj // subsampling] = radio[line]
                advancement.update()
            return sinogram

    @functools.lru_cache(maxsize=3)
    def getFlat(self, index: int | None = None):
        """
        If projectionI is not requested then return the mean value. Otherwise
        return the interpolated value for the requested projection.

        :param index: index of the projection for which we want the flat
        :return: Flat field value or None if can't deduce it
        """
        data = self._extractFromOneFile("refHST.edf", what="flat")
        if data is not None:
            return data

        data = self._extractFromPrefix(REFHST_PREFIX, what="flat", proI=index)
        if data is not None:
            return data

        _logger.warning("Cannot retrieve flat file from %s" % self.path)
        return None

    @docstring(TomwerScanBase.get_proj_angle_url)
    def get_proj_angle_url(self, use_cache: bool = True, *args, **kwargs):
        if not use_cache:
            self._cache_proj_urls = None

        if self._cache_proj_urls is None:
            self._cache_proj_urls = _tsEDFTomoScan.get_proj_angle_url(
                self, *args, **kwargs
            )
        return self._cache_proj_urls

    @functools.lru_cache()
    def getDark(self):
        """
        For now only deal with one existing dark file.

        :return: image of the dark if existing. Else None
        """
        if self._dark is None:
            # first try to retrieve data from dark.edf file or darkHST.edf files
            self._dark = self._extractFromOneFile("dark.edf", what="dark")
            if self._dark is None:
                self._dark = self._extractFromOneFile("darkHST.edf", what="dark")
            if self._dark is None:
                self._dark = self._extractFromPrefix(DARKHST_PREFIX, what="dark")
            if self._dark is None:
                self._dark = self._extractFromPrefix("darkend", what="dark")

            if self._dark is None:
                _logger.warning("Cannot retrieve dark file from %s" % self.path)

        return self._dark

    def _extractFromOneFile(self, f, what):
        if self.path is None:
            return None
        path = os.path.join(self.path, f)
        if os.path.exists(path):
            _logger.info(f"Getting {what} from {f}")
            try:
                data = fabio.open(path).data
            except Exception:
                return None
            else:
                if data.ndim == 2:
                    return data
                elif data.ndim == 3:
                    _logger.warning(
                        "%s file contains several images. Taking "
                        "the mean value" % what
                    )
                    return numpy.mean(data.ndim)
        else:
            return None

    @staticmethod
    def guess_index_frm_EDFFile_name(_file):
        name = _file
        if name.endswith(".edf"):
            name = name.rstrip(".edf")
        ic = []
        while name[-1].isdigit():
            ic.append(name[-1])
            name = name[:-1]

        if len(ic) == 0:
            return None
        else:
            orignalOrder = ic[::-1]
            return int("".join(orignalOrder))

    def _extractFromPrefix(self, pattern, what, proI=None):
        if self.path is None:
            return None
        files = glob(os.path.join(self.path, pattern + "*.edf"))
        if len(files) == 0:
            return None
        else:
            d = {}
            for f in files:
                index = self.guess_index_frm_EDFFile_name(f)
                if index is None:
                    _logger.error("cannot retrieve projection index for %s" "" % f)
                    return None
                else:
                    d[index] = fabio.open(f).data

            if len(files) == 1:
                return d[list(d.keys())[0]]

            oProj = OrderedDict(sorted(d.items()))
            # for now we only deal with interpolation between the higher
            # and the lower acquired file ()
            lowPI = list(oProj.keys())[0]
            uppPI = list(oProj.keys())[-1]

            lowPD = oProj[lowPI]
            uppPD = oProj[uppPI]

            if len(oProj) > 2:
                _logger.info(
                    f"Only bordering projections ({lowPI} and {uppPI}) will be used for extracting {what}"
                )

            uppPI = uppPI
            index = proI
            if index is None:
                index = (uppPI - lowPI) / 2

            if (index >= lowPI) is False:
                index = lowPI
                _logger.warning(
                    "ProjectionI not in the files indexes range (projectionI >= lowerProjIndex)"
                )

            if (index <= uppPI) is False:
                index = uppPI
                _logger.warning(
                    "ProjectionI not in the files indexes range upperProjIndex >= projectionI"
                )

            # simple interpolation
            _nRef = uppPI - lowPI
            lowPI = lowPI

            w0 = (lowPI + (uppPI - index)) / _nRef
            w1 = index / _nRef

            return w0 * lowPD + w1 * uppPD

    @staticmethod
    def get_index_reconstructed(reconstructionFile, scanID):
        """Return the slice reconstructed of a file from her name

        :param reconstructionFile: the name of the file
        """
        folderBasename = os.path.basename(scanID)
        if reconstructionFile.endswith(".edf") and reconstructionFile.startswith(
            folderBasename
        ):
            localstring = reconstructionFile.rstrip(".edf")
            # remove the scan
            localstring = re.sub(folderBasename, "", localstring)
            s = localstring.split("_")
            if s[-1].isdigit():
                return int(s[-1])
            else:
                _logger.warning(
                    "Fail to find the slice reconstructed for "
                    "file %s" % reconstructionFile
                )
        else:
            global counter_rand
            counter_rand = counter_rand + 1
            return counter_rand

    @docstring(_tsEDFTomoScan.to_dict)
    def to_dict(self):
        res = _tsEDFTomoScan.to_dict(self)
        res.update(TomwerScanBase.to_dict(self))
        return res

    @docstring(TomwerScanBase._deduce_transfert_scan)
    def _deduce_transfert_scan(self, output_dir):
        if os.path.basename(output_dir) != os.path.basename(self.path):
            raise ValueError(
                "Transfert to a new EDFTomoScan requires an equal basename. "
                f"Current path is {self.path}, requested one is {output_dir}"
            )
        # here: avoir reloading the metadata because not sure there will be some processing after the 'data transfert'. And can be long with EDF + GPFS at esrf
        return EDFTomoScan(scan=output_dir, update=False)

    def __str__(self):
        return self.path

    # Dataset implementation

    @docstring(_tsEDFTomoScan)
    def get_identifier(self):
        try:
            stat = pathlib.Path(self.path).stat()
        except Exception:
            stat = None
        return EDFTomoScanIdentifier(
            object=self,
            folder=self.path,
            file_prefix=self.dataset_basename,
            metadata={
                "name": self.path,
                "creation_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
                "modification_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
            },
        )

    @docstring(TomwerScanBase)
    def get_nabu_dataset_info(self, binning=1, binning_z=1, proj_subsampling=1):
        return {
            "hdf5_entry": "",
            "location": self.path,
            "binning": binning,
            "binning_z": binning_z,
            "projections_subsampling": proj_subsampling,
        }

    @docstring(TomwerScanBase)
    def to_nabu_dataset_analyser(self):
        from nabu.resources.dataset_analyzer import EDFDatasetAnalyzer

        analyzer = EDFDatasetAnalyzer(location=self.path)
        n_angles = analyzer.n_angles
        if analyzer.rotation_angles is None:
            if self.field_of_view is FOV.HALF:
                angles = numpy.linspace(0, 2 * numpy.pi, n_angles, True)
            else:
                angles = numpy.linspace(0, numpy.pi, n_angles, False)

            analyzer.rotation_angles = angles
        return analyzer

    @docstring(TomwerScanBase)
    def scan_dir_name(self) -> str | None:
        """for 'this/is/my/acquisition' returns 'acquisition'"""
        if self.path is not None:
            return self.path.split(os.sep)[-1]
        else:
            return None

    @docstring(TomwerScanBase)
    def scan_basename(self) -> str | None:
        """for 'this/is/my/acquisition' returns 'acquisition'"""
        if self.path is not None:
            return self.path
        else:
            return None

    @docstring(TomwerScanBase)
    def scan_parent_dir_basename(self) -> str | None:
        if self.path is not None:
            try:
                return os.path.dirname(self.path)
            except Exception:
                return None
        else:
            return None


def _get_urls_list(radio):
    """create the list of urls contained in a .edf file"""
    with fabio.open(radio) as edf_reader:
        if edf_reader.nframes == 1:
            return [DataUrl(file_path=radio, scheme="fabio")]
        else:
            res = []
            for iframe in range(edf_reader.nframes):
                res.append(
                    DataUrl(file_path=radio, data_slice=[iframe], scheme="fabio")
                )
            return res



from __future__ import annotations

import copy
import functools
import logging
import os
from typing import Iterable

from nabu import version as nabu_version
from nabu.pipeline.config import generate_nabu_configfile
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from processview.core.manager.manager import ProcessManager, DatasetState
from processview.core.superviseprocess import SuperviseProcess

from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as open_hdf5
from tomwer.core.utils.deprecation import deprecated_warning

from tomwer.core.cluster.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.process.reconstruction.nabu.plane import NabuPlane
from tomwer.core.process.task import Task
from tomwer.core.process.drac.processeddataset import DracReconstructedVolumeDataset
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.dictutils import concatenate_dict
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.io.utils.h5pyutils import EntryReader
from tomwer.utils import docstring
from tomwer.io.utils import format_stderr_stdout
from tomwer.core.process.reconstruction.output import (
    PROCESS_FOLDER_RECONSTRUCTED_VOLUMES,
)

from . import settings, utils
from .nabucommon import ResultsLocalRun, ResultSlurmRun, _NabuBaseReconstructor
from .target import Target

_logger = logging.getLogger(__name__)


def run_volume_reconstruction(
    scan: TomwerScanBase,
    config: dict,
    dry_run: bool,
    process_id: int | None = None,
) -> tuple:
    """
    Run a volume reconstruction. Scan need to have reconstruction parameters for nabu.

    Behavior: will clear link to the last volume reconstructed

    :param process_id: optional process id

    :return: succeed, stdouts, stderrs, configs, future_scan
    """

    if scan.nabu_recons_params in ({}, None):
        raise ValueError(
            "no configuration provided. You should run a "
            "reconstruction from nabuslices first."
        )
    cluster_config = config.pop("cluster_config", None)
    if cluster_config == {}:
        cluster_config = None
    elif isinstance(cluster_config, SlurmClusterConfiguration):
        cluster_config = cluster_config.to_dict()

    if cluster_config is None:
        target = Target.LOCAL
    else:
        target = Target.SLURM

    # # beam shape is not directly used by nabu (uses ctf_geometry directly)
    # config.get("phase", {}).pop("beam_shape", None)

    # config_volume = copy.copy(config)
    # config_nabu_slices = copy.deepcopy(scan.nabu_recons_params)
    # if "tomwer_slices" in config_nabu_slices:
    #     del config_nabu_slices["tomwer_slices"]

    # if "phase" in config_nabu_slices and "delta_beta" in config_nabu_slices["phase"]:
    #     pag_dbs = config_nabu_slices["phase"]["delta_beta"]
    #     if isinstance(pag_dbs, str):
    #         pag_dbs = utils.retrieve_lst_of_value_from_str(pag_dbs, type_=float)
    #     if len(pag_dbs) > 1:
    #         raise ValueError(
    #             "Several value of delta / beta found for volume reconstruction"
    #         )
    # scan.clear_latest_vol_reconstructions()

    if process_id is not None:
        try:
            process_name = ProcessManager().get_process(process_id=process_id).name
        except KeyError:
            process_name = "unknow"
    else:
        process_name = ""

    volume_reconstructor = VolumeRunner(
        scan=scan,
        config_nabu=config,
        cluster_config=cluster_config,
        dry_run=dry_run,
        target=target,
        process_name=process_name,
    )
    try:
        results = volume_reconstructor.run()
    except TimeoutError as e:
        _logger.error(e)
        return None
    else:
        assert len(results) == 1, "only one volume should be reconstructed"
        res = results[0]
        # tag latest reconstructions
        if isinstance(res, ResultsLocalRun) and res.results_identifiers is not None:
            scan.set_latest_vol_reconstructions(res.results_identifiers)
        # create future if needed
        if isinstance(res, ResultSlurmRun):
            future_tomo_obj = FutureTomwerObject(
                tomo_obj=scan,
                futures=tuple(res.future_slurm_jobs),
                process_requester_id=process_id,
            )

        else:
            future_tomo_obj = None
        succeed = res.success
        stdouts = (
            [
                res.std_out,
            ]
            if hasattr(res, "std_out")
            else []
        )
        stderrs = (
            [
                res.std_err,
            ]
            if hasattr(res, "std_err")
            else []
        )
        configs = (
            [
                res.config,
            ]
            if res is not None
            else []
        )

        return succeed, stdouts, stderrs, configs, future_tomo_obj


class VolumeRunner(_NabuBaseReconstructor):
    """
    Class used to reconstruct a full volume with Nabu.
    Locally or on a cluster.
    """

    EXPECTS_SINGLE_SLICE = False

    def __init__(
        self,
        scan: TomwerScanBase,
        config_nabu,
        cluster_config: dict | None,
        dry_run: bool,
        target: Target,
        process_name: str,
    ) -> None:
        super().__init__(
            scan=scan,
            dry_run=dry_run,
            target=target,
            cluster_config=cluster_config,
            process_name=process_name,
        )
        self._config = config_nabu

    @property
    def configuration(self):
        return self._config

    @property
    def processed_data_folder_name(self):
        """return the specific processed folder name associated to this type of reconstruction."""
        return PROCESS_FOLDER_RECONSTRUCTED_VOLUMES

    @docstring(_NabuBaseReconstructor)
    def run(self) -> Iterable:
        dataset_params = self.scan.get_nabu_dataset_info()
        if "dataset" in self.configuration:
            dataset_params.update(self.configuration["dataset"])
        self.configuration["dataset"] = dataset_params

        self.configuration["resources"] = self.configuration.get("resources", {})
        self.configuration["resources"]["method"] = "local"

        # force overwrite results
        if "output" not in self.configuration:
            self.configuration["output"] = {}
        config_slices, cfg_folder = self._treateOutputConfig(self.configuration)

        # force overwriting results
        config_slices["output"].update({"overwrite_results": 1})

        # check and clamp `start_z` and `end_z`
        if "reconstruction" in self.configuration:
            for key in ("start_z", "end_z"):
                value = config_slices["reconstruction"].get(key)
                if value is None:
                    continue

                value = int(value)
                if self.scan.dim_2 is not None and value >= self.scan.dim_2:
                    _logger.warning(
                        f"{key} > max_size (radio height: {self.scan.dim_2}). Set it to -1 (maximum)"
                    )
                    value = -1
                config_slices["reconstruction"][key] = value

        name = (
            config_slices["output"]["file_prefix"] + settings.NABU_CONFIG_FILE_EXTENSION
        )
        if not isinstance(self.scan, EDFTomoScan):
            name = "_".join((self.scan.entry.lstrip("/"), name))
        conf_file = os.path.join(cfg_folder, name)
        _logger.info(f"{self.scan}: create {conf_file}")

        # make sure output location exists
        os.makedirs(config_slices["output"]["location"], exist_ok=True)

        # add some tomwer metadata and save the configuration
        # note: for now the section is ignored by nabu but shouldn't stay that way
        with utils.TomwerInfo(config_slices) as config_to_dump:
            generate_nabu_configfile(
                conf_file,
                nabu_fullfield_default_config,
                config=config_to_dump,
                options_level="advanced",
            )
            return tuple(
                [
                    self._process_config(
                        config_to_dump=config_to_dump,
                        config_file=conf_file,
                        info="nabu volume reconstruction",
                        file_format=config_slices["output"]["file_format"],
                        process_name=self.process_name,
                    )
                ]
            )

    @docstring(_NabuBaseReconstructor)
    def _get_futures_slurm_callback(self, config_to_dump) -> tuple:
        # add callback to set slices reconstructed urls
        class CallBack:
            # we cannot create a future directly because distributed enforce
            # the callback to have a function signature with only the future
            # as single parameter.
            def __init__(self, f_partial, scan) -> None:
                self.f_partial = f_partial
                self.scan = scan

            def process(self, fn):
                if fn.done() and not (fn.cancelled() or fn.exception()):
                    # update reconstruction urls only if processing succeed.
                    recons_identifiers = self.f_partial()
                    self.scan.add_latest_vol_reconstructions(recons_identifiers)

        file_format = config_to_dump["output"]["file_format"]
        callback = functools.partial(
            utils.get_recons_volume_identifier,
            file_prefix=config_to_dump["output"]["file_prefix"],
            location=config_to_dump["output"]["location"],
            file_format=file_format,
            scan=self.scan,
            slice_index=None,
            axis=NabuPlane.XY,  # for volume we always reconstruct along XY plane
        )

        return (CallBack(callback, self.scan),)

    def _treateOutputConfig(self, config) -> tuple:
        """

        :return: (nabu config dict, nabu extra options)
        """
        config = copy.deepcopy(config)
        config, nabu_cfg_folder = super()._treateOutputSliceConfig(config)
        os.makedirs(config["output"]["location"], exist_ok=True)

        # adapt config_s to specific volume treatment
        if "postproc" in config:
            config["postproc"] = config["postproc"]

        # make sure start_[x] and end_[x] come from config
        for key in ("start_x", "end_x", "start_y", "end_y", "start_z", "end_z"):
            if key in config:
                config["reconstruction"][key] = config[key]
                del config[key]

        return config, nabu_cfg_folder

    @docstring(_NabuBaseReconstructor)
    def _get_file_basename_reconstruction(self, pag, db, ctf, axis):
        """

        :param pag: is it a Paganin reconstruction
        :param db: delta / beta parameter
        :param axis: axis over which the reconstruction goes. For volume always expected to be z. So ignored in the function
        :return: basename of the file reconstructed (without any extension)
        """
        assert type(db) in (int, type(None))
        assert not pag == ctf == True, "cannot ask for both pag and ctf active"
        if isinstance(self.scan, NXtomoScan):
            basename, _ = os.path.splitext(self.scan.master_file)
            basename = os.path.basename(basename)
            try:
                # if there is more than one entry in the file append the entry name to the file basename
                with open_hdf5(self.scan.master_file) as h5f:
                    if len(h5f.keys()) > 1:
                        basename = "_".join((basename, self.scan.entry.strip("/")))
            except Exception:
                pass
        else:
            basename = os.path.basename(self.scan.path)

        if pag:
            return "_".join((basename + "pag", "db" + str(db).zfill(4), "vol"))
        elif ctf:
            return "_".join((basename + "ctf", "db" + str(db).zfill(4), "vol"))
        else:
            return "_".join((basename, "vol"))


class NabuVolumeTask(
    Task,
    SuperviseProcess,
    input_names=("data", "nabu_params"),
    output_names=(
        "data",
        "volumes",
        "future_tomo_obj",
        "data_portal_processed_datasets",
    ),
    optional_input_names=(
        "dry_run",
        "nabu_extra_params",  # some parameter that must update 'nabu_params' before launching the reconstruction. Such as z range...
        "serialize_output_data",
    ),
):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        SuperviseProcess.__init__(self, process_id=process_id)
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        self._dry_run = inputs.get("dry_run", False)
        self._current_processing = None

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            self.outputs.data = None
            return

        # update scan reconstruction parameters used
        nabu_params = concatenate_dict(
            copy.deepcopy(self.inputs.nabu_params),
            self.get_input_value("nabu_extra_params", dict()),
        )
        scan.nabu_recons_params = nabu_params
        scan.clear_latest_vol_reconstructions()

        cluster_config = nabu_params.pop("cluster_config", None)
        if cluster_config == {}:
            cluster_config = None
        elif isinstance(cluster_config, SlurmClusterConfiguration):
            cluster_config = cluster_config.to_dict()

        if "tomwer_slices" in nabu_params:
            del nabu_params["tomwer_slices"]

        if "phase" in nabu_params and "delta_beta" in nabu_params["phase"]:
            pag_dbs = nabu_params["phase"]["delta_beta"]
            if isinstance(pag_dbs, str):
                pag_dbs = utils.retrieve_lst_of_value_from_str(pag_dbs, type_=float)
            if len(pag_dbs) > 1:
                raise ValueError(
                    "Several value of delta / beta found for volume reconstruction"
                )

        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(f"input type {scan} is not managed")

        if cluster_config is None:
            target = Target.LOCAL
        else:
            target = Target.SLURM

        state = None
        details = None
        try:
            self._current_processing = VolumeRunner(
                scan=scan,
                config_nabu=nabu_params,
                cluster_config=cluster_config,
                dry_run=self._dry_run,
                target=target,
                process_name=self.program_name,
            )
            res = self._current_processing.run()[0]
        except Exception as e:
            _logger.error(f"Failed to process {scan}. Error is {e}")
            state = DatasetState.FAILED
            details = None
            future_tomo_obj = None
        else:
            # tag latest reconstructions
            if isinstance(res, ResultsLocalRun) and res.results_identifiers is not None:
                scan.set_latest_vol_reconstructions(res.results_identifiers)
            # create future if needed
            if isinstance(res, ResultSlurmRun):
                future_tomo_obj = FutureTomwerObject(
                    tomo_obj=scan,
                    futures=tuple(res.future_slurm_jobs),
                    process_requester_id=self.process_id,
                )

            else:
                future_tomo_obj = None

            if self._cancelled:
                state = DatasetState.CANCELLED
                details = "cancelled by user"
                _logger.info(f"Slices computation for {scan} cancelled")
            else:
                succeed = res.success
                stdouts = (
                    [
                        res.std_out,
                    ]
                    if hasattr(res, "std_out")
                    else []
                )
                stderrs = (
                    [
                        res.std_err,
                    ]
                    if hasattr(res, "std_err")
                    else []
                )

                if not succeed:
                    mess = f"Volume computed for {scan} failed."
                    _logger.processFailed(mess)
                    state = DatasetState.FAILED
                else:
                    mess = f"Volume computed for {scan}."
                    _logger.processSucceed(mess)
                    state = DatasetState.SUCCEED

                # format stderr and stdout
                elmts = [
                    format_stderr_stdout(stderr=stderr, stdout=stdout)
                    for stderr, stdout in zip(stderrs, stdouts)
                ]
                elmts.insert(0, mess)
                details = "\n".join(elmts)
        finally:
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self,
                state=state,
                details=details,
            )

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan
        self.outputs.volumes = scan.latest_vol_reconstructions
        self.outputs.future_tomo_obj = future_tomo_obj

        # build screenshots
        data_portal_processed_datasets = []
        if scan.latest_vol_reconstructions is not None:
            for volume_id in scan.latest_vol_reconstructions:
                try:
                    volume = VolumeFactory.create_tomo_object_from_identifier(
                        identifier=volume_id
                    )
                except Exception as e:
                    _logger.error(
                        f"Fail to build volume from {volume_id}. Error is {e}"
                    )
                else:
                    icatReconstructedDataset = DracReconstructedVolumeDataset(
                        tomo_obj=volume,
                        source_scan=scan,
                    )
                    data_portal_processed_datasets.append(icatReconstructedDataset)

        self.outputs.data_portal_processed_datasets = tuple(
            data_portal_processed_datasets
        )

    def set_configuration(self, configuration: dict) -> None:
        Task.set_configuration(self, configuration=configuration)
        if "dry_run" in configuration:
            self.set_dry_run(bool(configuration["dry_run"]))

    @staticmethod
    def program_name():
        return "nabu-volume"

    @staticmethod
    def program_version():
        return nabu_version

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @staticmethod
    def retrieve_last_relative_cor(scan):
        with EntryReader(scan.process_file_url) as h5f:
            latest_nabu_node = Task.get_most_recent_process(h5f, NabuVolumeTask)
            path = "configuration/reconstruction/rotation_axis_position"
            if latest_nabu_node is not None and path in latest_nabu_node:
                return h5py_read_dataset(latest_nabu_node[path])

    def cancel(self):
        """
        stop current processing
        """
        self._cancelled = True
        if self._current_processing is not None:
            self._current_processing.cancel()


class NabuVolume(NabuVolumeTask):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        deprecated_warning(
            name="tomwer.core.process.reconstruction.nabu.nabuvolume.NabuVolume",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="NabuVolumeTask",
        )
        super().__init__(process_id, varinfo, inputs, node_id, node_attrs, execinfo)



# coding: utf-8

"""
This module is used to manage observations. Initially on files.
Observations are runned on a thread and run each n seconds.
They are manage by thread and signals
"""


import threading
import time


class LoopThread(threading.Thread):
    """
    Thread looping on a defined process, as long not receiving the stopEvent.
    Wait for `_time_between_loops` between two loops
    """

    startEvent = threading.Event()
    """Event to restart the loop on process"""
    stopEvent = threading.Event()
    """Event to stop the loop on process"""
    quitEvent = threading.Event()
    """Event to quit the Thread"""

    def __init__(self, time_between_loops):
        threading.Thread.__init__(self)
        self._time_between_loops = time_between_loops
        self._status = None
        self._last_processing = None

    def run(self):
        threading.Thread.run(self)
        while not self.quitEvent.isSet():
            if self._status is None or (
                self._status == "not running" and self.startEvent.isSet()
            ):
                self._status = "running"
                self.startEvent.clear()

            if self._status == "running" and self.stopEvent.isSet() is True:
                self.stopEvent.clear()
                self._status = "not running"
            if self._status == "running" and self.should_process():
                self._last_processing = time.time()
                self.process()
            time.sleep(0.05)

    def process(self):
        raise NotImplementedError("Pure virtual class")

    def should_process(self):
        if self._last_processing is None:
            return True
        _pass_time_since_last_pro = time.time() - self._last_processing
        return _pass_time_since_last_pro > self._time_between_loops

    def _stop(self):
        pass



"""Contains the QImageFileStackPlot. Widget to display a set of files with metadata (either scan or volume metadata)
"""

from __future__ import annotations

import logging
import os
import time

from silx.gui import qt
from tomwer.gui.visualization.imagestack import ImageStack


from . import utils

logger = logging.getLogger(__name__)


class QImageFileStackPlot(ImageStack):
    """
    Widget based on ImageStack and adding some widget to display metadata associated with urls.
    Like Paganin delta / beta values...
    """

    def __init__(self, parent, show_overview=True, backend=None):
        """
        Constructor

        :param parent: the Qt parent widget
        """
        super().__init__(parent, show_overview=show_overview)

        self._plot.setBackend(backend=backend)

        self._imagejThreads = []

        # dock widget: qslider
        self._plot.layout()
        self._fileInfoDockWidget = qt.QDockWidget(parent=self)
        self._fileInfoWidget = self.__buildFileInfo()
        self._fileInfoDockWidget.setWidget(self._fileInfoWidget)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._fileInfoDockWidget)
        self._fileInfoDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)

        # connect signal / slot
        self._openWithImjButton.released.connect(self._openCurrentInImagej)
        self.sigCurrentUrlChanged.connect(self._updateUrlInfos)

    def __buildFileInfo(self):
        self._fileInfoWidget = qt.QWidget(parent=self)
        layout = qt.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self._fileInfoWidget.setLayout(layout)

        # file name label
        layout.addWidget(qt.QLabel("file :", self._fileInfoWidget), 0, 0, 1, 1)
        self._qlFileName = qt.QLabel("", parent=self._fileInfoWidget)
        layout.addWidget(self._qlFileName, 0, 1, 1, 2)

        # open in image j button
        style = qt.QApplication.style()
        open_icon = style.standardIcon(qt.QStyle.SP_FileLinkIcon)
        self._openWithImjButton = qt.QPushButton(
            open_icon, "open with ImageJ", parent=self
        )
        self._openWithImjButton.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        layout.addWidget(self._openWithImjButton, 0, 3, 1, 1)

        # date last modification
        layout.addWidget(
            qt.QLabel("last modification :", self._fileInfoWidget), 1, 0, 1, 1
        )
        self._qlLastModifications = qt.QLabel("", parent=self._fileInfoWidget)
        layout.addWidget(self._qlLastModifications, 1, 1, 1, 2)

        return self._fileInfoWidget

    def _openCurrentInImagej(self):
        current_url = self.getCurrentUrl()
        if current_url is None:
            logger.warning("No active image defined")
        else:
            try:
                self._imagejThreads.append(utils.open_url_with_image_j(current_url))
            except OSError as e:
                msg = qt.QMessageBox(self)
                msg.setIcon(qt.QMessageBox.Warning)
                msg.setWindowTitle("Unable to open image in imagej")
                msg.setText(str(e))
                msg.exec_()

    def _updateUrlInfos(self):
        url = self.getCurrentUrl()
        name = url.file_path().split(os.path.sep)[-1]

        self._qlFileName.setText(name)
        self._qlFileName.setToolTip(url.path())
        try:
            last_mod = time.ctime(os.path.getmtime(url.file_path()))
        except Exception:
            last_mod = ""
        self._qlLastModifications.setText(last_mod)



# coding: utf-8
from __future__ import annotations

import getpass
import logging
import socket
import threading
import time
from multiprocessing import Process

from jsonrpc import JSONRPCResponseManager, dispatcher
from psutil import process_iter
from silx.utils.enum import Enum as _Enum
from werkzeug.serving import run_simple
from werkzeug.wrappers import Request, Response

_logger = logging.getLogger(__name__)


class TangoAcquisitionStatus(_Enum):
    """list of possible status of the acquisition"""

    STARTED = "started"
    ON_GOING = "on going"
    CANCELED = "canceled"
    ENDED = "ended"


class _TangoState:
    MOVING = "moving"
    ON = "on"
    FAULT = "fault"


class BlissAcquisition:
    """Define an acquisition made with bliss / tango"""

    def __init__(self, file_path, entry_name, proposal_file, sample_file, start_time):
        self.entry = entry_name
        self.master_file = file_path
        self.proposal_file = proposal_file
        self.sample_file = sample_file
        self.scan_numbers = []
        self.status = TangoAcquisitionStatus.STARTED
        self._start_time = start_time
        self._end_time = None
        self._error = None
        self._state = None

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        self._state = state

    def add_scan_number(self, scan_number):
        self.scan_numbers.append(scan_number)

    def set_status(self, status):
        self.status = TangoAcquisitionStatus.from_value(status)

    def end(self, end_time, succeed: bool, error: str | None):
        if succeed is True:
            self.status = TangoAcquisitionStatus.ENDED
        else:
            self.status = TangoAcquisitionStatus.CANCELED
        self._error = error
        self._end_time = end_time

    @property
    def has_error(self):
        return self._error is not None

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time

    @property
    def error(self) -> str | None:
        return self.error


class _BaseDataListenerThread:
    """Base class for data listener thread. Thread can be a threading.Thread
    or a qt.QThread.

    On the current bliss system (2020) a sequence is an acquisition. A sequence
    is composed of several scans. A scan can be the 'init' information,
    dark frames, flat frames, projections frames, return projections...

    """

    _JOIN_TIMEOUT = 5  # s

    def __init__(self, host, port, acquisitions=None):
        assert host is not None
        assert port is not None
        # connect to the tango device
        self.acquisitions = acquisitions or []
        """acquisition with scan information, as key and acquisition status
        as value"""
        self._host = host
        # if host is None then use hostname
        if self._host is None:
            self._host = socket.gethostname()
        self._port = port
        self._tomo_state = None
        # state of the current sequence
        self._current_sequence = None
        # current sequence
        self._current_scan_number = None
        # scan of the sequence currently recording
        self._server = None
        self._stop = False

    def _rpc_sequence_started(
        self, saving_file, scan_title, sequence_scan_number, proposal_file, sample_file
    ):
        self._current_sequence = BlissAcquisition(
            file_path=saving_file,
            entry_name=sequence_scan_number,
            proposal_file=proposal_file,
            start_time=time.ctime(),
            sample_file=sample_file,
        )
        self.sequence_started(acquisition=self._current_sequence)

    def sequence_started(self, acquisition: BlissAcquisition):
        print(
            "sequence {}@{} started. Title is {}, proposal file {}"
            "".format(
                acquisition.entry,
                acquisition.master_file,
                acquisition.title,
                acquisition.proposal_file,
            )
        )

    def _rpc_scan_started(self, scan_number):
        if self._current_sequence is not None:
            self._current_sequence.add_scan_number(scan_number=scan_number)
            self.scan_added(acquisition=self._current_sequence, scan_number=scan_number)

    def scan_added(self, acquisition: BlissAcquisition, scan_number: int):
        print(f"scan {scan_number} started")

    def _rpc_scan_ended(self, scan_number):
        if self._current_sequence is not None:
            self.scan_ended(acquisition=self._current_sequence, scan_number=scan_number)

    def scan_ended(self, acquisition: BlissAcquisition, scan_number):
        print(f"scan {scan_number} ended")

    def _rpc_sequence_ended(self, saving_file, sequence_scan_number, success: bool):
        if success is False:
            _logger.warning(f"sequence {sequence_scan_number}@{saving_file} failed")
            if self._current_sequence is not None:
                try:
                    self.sequence_failed(self._current_sequence)
                except Exception as e:
                    _logger.warning(f"Fail to catch cancel scan. error is {e}")

            return

        if self._current_sequence is not None:
            if (
                self._current_sequence.master_file == saving_file
                and self._current_sequence.entry == sequence_scan_number
            ):
                self.sequence_ended(self._current_sequence)
            else:
                name = "@".join((sequence_scan_number, saving_file))
                name_current = "@".join(
                    (self._current_sequence.entry, self._current_sequence.master_file)
                )

                _logger.warning(
                    f"End of {name} detected but does not fit the current acquisition {name_current}"
                )

    def sequence_ended(self, acquisition: BlissAcquisition):
        """callback when a sequence end (with success)"""
        print(
            f"sequence {acquisition.entry}@{acquisition.master_file} ended. Succeed: {acquisition.status}"
        )

    def sequence_failed(self, acquisition: BlissAcquisition):
        """callback when a sequence fails (probably canceled)"""
        print(f"sequence {acquisition.entry}@{acquisition.master_file} failed")

    @Request.application
    def application(self, request):
        dispatcher["scan_started"] = self._rpc_scan_started
        dispatcher["scan_ended"] = self._rpc_scan_ended
        dispatcher["sequence_started"] = self._rpc_sequence_started
        dispatcher["sequence_ended"] = self._rpc_sequence_ended

        response = JSONRPCResponseManager.handle(request.data, dispatcher)
        return Response(response.json, mimetype="application/json")

    def start(self) -> None:
        self._stop = False
        _logger.info(f"launch rpc server on {self._host}:{self._port}")
        self._server = Process(
            target=run_simple, args=(self._host, self._port, self.application)
        )
        self._server.start()
        _logger.info(f"launcher server with pid {self._server.pid}")

    def stop(self):
        self._stop = True
        if self._server is not None:
            _logger.info(f"stop rpc server on {self._host}:{self._port}")
            self._server.terminate()
            self._server.join(_BaseDataListenerThread._JOIN_TIMEOUT)
            self._server = None


class DataListenerThread(_BaseDataListenerThread, threading.Thread):
    """Implementation of _BaseDataListenerThread with a threading.Thread"""

    def __init__(self, host, port, acquisitions=None):
        threading.Thread.__init__(self)
        _BaseDataListenerThread.__init__(
            self, host=host, port=port, acquisitions=acquisitions
        )


def send_signal_to_local_rpc_servers(signal, port: int, extended_find: bool = True):
    """
    :param signal: signal to be emit
    :param port: port to check
    :param extended_find: if True then will try to find a process that occupy the port
        even if this is not a process launched by the user and launched by tomwer.
    """
    # TODO: this should be moved to some utils
    import psutil

    found = False
    for proc in process_iter():
        # try to find a process we can handle
        if proc.username() == getpass.getuser():
            try:
                for conns in proc.connections():
                    # make sure we will kill the correct process
                    if conns.laddr.port == port and proc.name() in (
                        "tomwer",
                        "orange-canvas",
                    ):
                        _logger.warning(f"send {signal} signal to pid {proc.pid}")
                        proc.send_signal(signal)
                        found = True
                        return
            except (psutil.PermissionError, psutil.AccessDenied):
                pass
    if not extended_find:
        return
    # if process not found try to find one to inform the user
    if not found:
        for proc in process_iter():
            try:
                for conns in proc.connections():
                    # make sure we will kill the correct process
                    if conns.laddr.port == port:
                        _logger.warning(
                            f"process pid: {proc.pid} - {proc.name()} seems to be one occupying port {port}"
                        )
                        return
            except (psutil.PermissionError, psutil.AccessDenied):
                pass



from __future__ import annotations

from silx.gui import qt
from nabu.stitching.alignment import AlignmentAxis1, AlignmentAxis2


class _AlignmentGroupBox(qt.QGroupBox):
    DEFAULT_PAD_MODE = "constant"

    ALIGNMENT_DOC = (
        "https://tomotools.gitlab-pages.esrf.fr/nabu/stitching/alignment.html"
    )

    DEFAULT_ALIGNMENT_AXIS_1 = AlignmentAxis1.CENTER
    DEFAULT_ALIGNMENT_AXIS_2 = AlignmentAxis2.CENTER

    _PAD_MODES = (
        "constant",  # Pads with a constant value.
        "edge",  # Pads with the edge values of array.
        "linear_ramp",  # Pads with the linear ramp between end_value and the array edge value.
        "maximum",  # Pads with the maximum value of all or part of the vector along each axis.
        "mean",  # Pads with the mean value of all or part of the vector along each axis.
        "median",  # Pads with the median value of all or part of the vector along each axis.
        "minimum",  # Pads with the minimum value of all or part of the vector along each axis.
        "reflect",  # Pads with the reflection of the vector mirrored on the first and last values of the vector along each axis.
        "symmetric",  # Pads with the reflection of the vector mirrored along the edge of the array.
        "wrap",  # Pads with the wrap of the vector along the axis. The first values are used to pad the end and the end values are used to pad the beginning.
    )

    def __init__(self, parent: qt.QWidget = None, title="alignment") -> None:
        super().__init__(title, parent)
        self.setLayout(qt.QFormLayout())

        # alignment axis 1
        self._alignmentAxis1CB = qt.QComboBox(self)
        for alignment in AlignmentAxis1.values():
            self._alignmentAxis1CB.addItem(alignment)
        self.layout().addRow("Axis 1 alignment", self._alignmentAxis1CB)
        self._alignmentAxis1CB.setToolTip(
            f"Alignment to do in case of volumes with different size over axis 1. Only possible for post-processing (reconstructed volume). See {self.ALIGNMENT_DOC} for details."
        )

        # alignment axis 2
        self._alignmentAxis2CB = qt.QComboBox(self)
        for alignment in AlignmentAxis2.values():
            self._alignmentAxis2CB.addItem(alignment)
        self.layout().addRow("Axis 2 alignment", self._alignmentAxis2CB)
        self._alignmentAxis2CB.setToolTip(
            f"Alignment to do in case of frames with different size over axis 2. See {self.ALIGNMENT_DOC} for details."
        )

        # pad mode
        self._padModeCB = qt.QComboBox(self)
        for pad_mode in self._PAD_MODES:
            self._padModeCB.addItem(pad_mode)
        self.layout().addRow("pad mode", self._padModeCB)
        self._padModeCB.setToolTip("padding mode to apply for alignment")

        # set up
        self.setAlignmentAxis1(self.DEFAULT_ALIGNMENT_AXIS_1)
        self.setAlignmentAxis2(self.DEFAULT_ALIGNMENT_AXIS_2)

    def getAlignmentAxis1(self) -> AlignmentAxis1:
        return AlignmentAxis1.from_value(self._alignmentAxis1CB.currentText())

    def setAlignmentAxis1(self, alignment: AlignmentAxis1):
        alignment = AlignmentAxis1.from_value(alignment)
        self._alignmentAxis1CB.setCurrentIndex(
            self._alignmentAxis1CB.findText(alignment.value)
        )

    def getAlignmentAxis2(self) -> AlignmentAxis2:
        return AlignmentAxis2.from_value(self._alignmentAxis2CB.currentText())

    def setAlignmentAxis2(self, alignment: AlignmentAxis2):
        alignment = AlignmentAxis2.from_value(alignment)
        self._alignmentAxis2CB.setCurrentIndex(
            self._alignmentAxis2CB.findText(alignment.value)
        )

    def getPadMode(self) -> str:
        return self._padModeCB.currentText()

    def setPadMode(self, pad_mode: str):
        idx = self._padModeCB.findText(pad_mode)
        if idx >= 0:
            self._padModeCB.setCurrentIndex(idx)

    def setAlignmentAxis1Enabled(self, enabled: bool):
        self._alignmentAxis1CB.setEnabled(enabled)



# coding: utf-8

"""
This module is used to define a set of folders to be emitted to the next box.
"""

import logging

from tomwer.core.process.task import Task
from tomwer.core.utils.scanutils import data_identifier_to_scan

logger = logging.getLogger(__name__)


class _ScanListPlaceHolder(
    Task, optional_input_names=("data",), output_names=("data",)
):
    """For now data can only be a single element and not a list.
    This must be looked at.
    Also when part of an ewoks graph 'data' is mandatory which is not the class
    when part of a orange workflow. Those can be added interactively"""

    def run(self):
        self.outputs.data = data_identifier_to_scan(self.inputs.data)



# coding: utf-8

"""some utils relative to PyHST
"""

from __future__ import annotations

import numpy
from silx.gui import qt


class LogSlider(qt.QWidget):
    """Slider to select a value with a QSlider displayed with log scale"""

    valueChanged = qt.Signal(float)
    """signal emitted when the value changed"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        # QSlider
        self._slider = qt.QSlider(self)
        self._slider.setOrientation(qt.Qt.Horizontal)
        self.layout().addWidget(self._slider, 0, 0, 1, 1)
        # Double spin box
        self._valueQBSB = qt.QDoubleSpinBox(self)
        self.layout().addWidget(self._valueQBSB, 0, 1, 1, 1)

        # connect signal / slot
        self._slider.valueChanged.connect(self._sliderValueChanged)
        self._valueQBSB.valueChanged.connect(self._qdsbValueChanged)
        # set up
        self.setRange(1, 100)
        self.setValue(5)

    def setSuffix(self, txt):
        self._valueQBSB.setSuffix(txt)

    def setPrefix(self, txt):
        self._valueQBSB.setPrefix(txt)

    def setRange(self, min_: float, max_: float) -> None:
        """
        Define slider range

        :param min_:
        :param max_:
        """
        if min_ <= 0.0 or max_ <= 0.0:
            raise ValueError("LogSlider can only handled positive values")
        self._valueQBSB.setRange(min_, max_)
        self._slider.setRange(int(numpy.log(min_)), int(numpy.log(max_)))

    def _sliderValueChanged(self, *args, **kwargs):
        old = self._valueQBSB.blockSignals(True)
        self._valueQBSB.setValue(numpy.exp(self._slider.value()))
        self._valueQBSB.blockSignals(old)
        self.valueChanged.emit(self.value())

    def _qdsbValueChanged(self, *args, **kwargs):
        old = self._slider.blockSignals(True)
        self._slider.setValue(int(numpy.log(self._valueQBSB.value())))
        self._slider.blockSignals(old)
        self.valueChanged.emit(self.value())

    def value(self):
        return self._valueQBSB.value()

    def setValue(self, value):
        self._valueQBSB.setValue(value)



from __future__ import annotations


import functools
import numpy
from silx.gui import qt
from silx.utils.enum import Enum as _Enum

from tomwer.core.tomwer_object import TomwerObject
from tomwer.gui import icons
from tomwer.gui.stitching.axisorderedlist import AxisOrderedTomoObjsModel
from tomwer.gui.utils.step import StepSizeSelectorWidget
from tomwer.gui.utils.qt_utils import block_signals


class PosEditorOverOneAxis(qt.QWidget):
    """keep it ordered along one axis"""

    def __init__(
        self, parent, axis_edited: int, axis_order=None, *args, **kwargs
    ) -> None:
        assert axis_edited in (0, 1, 2)
        super().__init__(parent, *args, **kwargs)
        self._axisEdited = axis_edited
        # the axis we are editing
        self._axisOrder = axis_order if axis_order is not None else axis_edited
        # the axis along which the tomo obj are ordered
        self._tomoObjtoTomoObjPosWidget = {}
        # list of `_TomoObjPosition` to edit the position over the axis. Key is the tomo object, value is the `_TomoObjPosition`
        self.__tomoObjPosWidgetCallbacks = {}
        self.setLayout(qt.QVBoxLayout())
        # widget to define step size
        self._stepSizeWidget = StepSizeSelectorWidget(
            self,
            fine_value=1,
            medium_value=5,
            rough_value=25,
            dtype=int,
        )
        self.layout().addWidget(self._stepSizeWidget)

        # widget to define edition mode and reset
        self._editionOptionsWidget = _EditionOptions()
        self.layout().addWidget(self._editionOptionsWidget)

        # table with the different Tomo objects
        self._tomoObjsTableView = qt.QTableView(parent=self)
        model = EditableAxisOrderedTomoObjsModel(axis=axis_edited)
        self._tomoObjsTableView.setModel(model)

        self.layout().addWidget(self._tomoObjsTableView)

        # connect signal / slot
        self._stepSizeWidget.valueChanged.connect(self._updateStepSize)
        self._editionOptionsWidget.sigReset.connect(self._resetPositions)

        # tune table view
        self._tomoObjsTableView.setColumnWidth(0, 45)
        self._tomoObjsTableView.setColumnWidth(2, _TomoObjPosition.WIDGET_WIDTH)
        self._tomoObjsTableView.horizontalHeader().setSectionResizeMode(
            1, qt.QHeaderView.Stretch
        )
        self.setStepSize(1)

    def _updateStepSize(self):
        step_size = self.getStepSize()
        for sb in self._tomoObjtoTomoObjPosWidget.values():
            sb.setSingleStep(step_size)

    def setStepSize(self, step_size: int):
        self._stepSizeWidget.setStepSize(step_size)
        self._updateStepSize()

    def getStepSize(self) -> int:
        return self._stepSizeWidget.getStepSize()

    def getTomoObjs(self) -> tuple[TomwerObject]:
        return self._tomoObjsTableView.model().getTomoObjs()

    def addTomoObj(self, tomo_obj):
        if tomo_obj is None:
            return
        else:
            self._tomoObjsTableView.model().addTomoObj(tomo_obj)
            # register tomo obj metadata modification to make sure we keel the z ordered list up to data
            tomo_obj.stitching_metadata.sigChanged.connect(
                self._orderedMightHaveChanged
            )
            self._createTomoObjPosition(tomo_obj=tomo_obj)
            self._orderedMightHaveChanged()

    def _createTomoObjPosition(self, tomo_obj, original_value: float | None = None):
        """
        :param original_value: original_value of the tomo_obj along the edited axis. Used when the widget is delete and recreate
        """
        widget = _TomoObjPosition(
            tomo_obj=tomo_obj,
            axis_edited=self._axisEdited,
            original_value=original_value,
            parent=self,
        )
        widget.setSingleStep(self.getStepSize())

        identifier_as_str = tomo_obj.get_identifier().to_str()
        self._tomoObjtoTomoObjPosWidget[identifier_as_str] = widget
        # connect signal / slot
        callback = functools.partial(
            self._tomoObjPosChanged, spin_box=widget, tomo_obj_to_update=tomo_obj
        )
        self.__tomoObjPosWidgetCallbacks[identifier_as_str] = callback
        widget.sigValueChanged.connect(callback)
        return widget

    def _deleteSpinBox(self, tomo_obj):
        identifier_as_str = tomo_obj.get_identifier().to_str()
        if identifier_as_str in self._tomoObjtoTomoObjPosWidget:
            spinBox = self._tomoObjtoTomoObjPosWidget.pop(identifier_as_str)
            spinBox.deleteLater()
        if identifier_as_str in self.__tomoObjPosWidgetCallbacks:
            del self.__tomoObjPosWidgetCallbacks[identifier_as_str]

    def removeTomoObj(self, tomo_obj):
        self._deleteSpinBox(tomo_obj)
        self._tomoObjsTableView.model().removeTomoObj(tomo_obj)
        tomo_obj.stitching_metadata.sigChanged.disconnect(self._orderedMightHaveChanged)

    def _tomoObjPosChanged(
        self,
        old_pos_value: int,
        new_pos_value: int,
        tomo_obj_to_update,
        *args,
        **kwargs,
    ):
        tomo_obj_to_update.stitching_metadata.setPxPos(
            new_pos_value, axis=self._axisEdited
        )
        # edit upstream or downstream nodes according to the edit mode
        shift = new_pos_value - old_pos_value
        if shift == 0:
            return

        edition_mode = self._editionOptionsWidget.getEditionMode()
        if edition_mode is _EditionMode.FREE:
            tomo_objs_to_update = ()
        elif edition_mode is _EditionMode.DOWNSTREAM:
            tomo_objs_to_update = self._get_downstream_tomo_obj(
                position=old_pos_value, tomo_obj_to_filter=tomo_obj_to_update
            )
        elif edition_mode is _EditionMode.UPSTREAM:
            tomo_objs_to_update = self._get_upstream_tomo_obj_pos_widgets(
                position=old_pos_value, tomo_obj_to_filter=tomo_obj_to_update
            )
        else:
            raise NotImplementedError(f"edition mode ({edition_mode}) is not handled")
        with block_signals(self):
            for tomo_obj_to_update in tomo_objs_to_update:
                associated_widget = self._tomoObjtoTomoObjPosWidget[
                    tomo_obj_to_update.get_identifier().to_str()
                ]
                current_value = associated_widget.getValue()
                assert (
                    current_value == associated_widget.getValue()
                ), "incoherent value between tomo object and widget"
                new_value = current_value + shift
                with block_signals(associated_widget):
                    associated_widget.setValue(new_value)
                with block_signals(tomo_obj_to_update.stitching_metadata):
                    tomo_obj_to_update.stitching_metadata.setPxPos(
                        value=new_value, axis=self._axisEdited
                    )
        self._orderedMightHaveChanged()

    def setTomoObjs(self, tomo_objs: tuple) -> None:
        """
        replace current list of object by the given list
        """
        self._tomoObjsTableView.model().clearTomoObjs()
        for tomo_obj in tomo_objs:
            self.addTomoObj(tomo_obj)

    def clean(self):
        tomo_objs = self._tomoObjsTableView.model().getTomoObjs()
        for tomo_obj in tomo_objs:
            self.removeTomoObj(tomo_obj=tomo_obj)

    def setSeries(self):
        raise NotImplementedError

    def _resetPositions(self):
        tomo_objs = self.getTomoObjs()

        with block_signals(self):
            for tomo_obj in tomo_objs:
                widget = self._tomoObjtoTomoObjPosWidget[
                    tomo_obj.get_identifier().to_str()
                ]
                with block_signals(widget):
                    widget.resetOriginalValue()
                    tomo_obj.stitching_metadata.setPxPos(
                        value=widget.getOriginalValue(), axis=self._axisEdited
                    )
        self._orderedMightHaveChanged()

    def _get_upstream_tomo_obj_pos_widgets(
        self, position: float, tomo_obj_to_filter: TomwerObject | None = None
    ) -> tuple[TomwerObject]:
        return tuple(
            filter(
                lambda tomo_obj: (
                    tomo_obj.stitching_metadata.get_abs_position_px(
                        axis=self._axisEdited
                    )
                    >= position
                )
                and (tomo_obj is not tomo_obj_to_filter),
                self.getTomoObjs(),
            )
        )

    def _get_downstream_tomo_obj(
        self, position: float, tomo_obj_to_filter: TomwerObject | None = None
    ) -> tuple[TomwerObject]:
        return tuple(
            filter(
                lambda tomo_obj: (
                    tomo_obj.stitching_metadata.get_abs_position_px(
                        axis=self._axisEdited
                    )
                    <= position
                )
                and (tomo_obj is not tomo_obj_to_filter),
                self.getTomoObjs(),
            )
        )

    def _orderedMightHaveChanged(self, force_sb_update=False):
        # add index widget
        self._tomoObjsTableView.model().reorder_objs()
        self._tomoObjsTableView.model().layoutChanged.emit()

        ordered_objs = self._tomoObjsTableView.model()._axis_decreasing_ordered_objs

        # start update position widgets
        # check if we need to update one widget. Has this is designed if we need to update one then we need to update them all...
        needs_to_update_widget = force_sb_update
        for i_pos, tomo_obj in enumerate(ordered_objs):
            identifier_as_str = tomo_obj.get_identifier().to_str()
            widget = self._tomoObjtoTomoObjPosWidget[identifier_as_str]
            model_index = self._tomoObjsTableView.model().createIndex(i_pos, 2)
            if self._tomoObjsTableView.indexWidget(model_index) not in (None, widget):
                needs_to_update_widget = True
                break

        for i_pos, tomo_obj in enumerate(ordered_objs):
            identifier_as_str = tomo_obj.get_identifier().to_str()
            widget = self._tomoObjtoTomoObjPosWidget[identifier_as_str]
            model_index = self._tomoObjsTableView.model().createIndex(i_pos, 2)
            if needs_to_update_widget:
                # if items have been reordered then we must recreated `TomoObjPosWidget` otherwise if we try
                # to change order then Qt will end up with a seg fault which seems to come from
                # overwriting the cell and trying to reuse them
                original_value = widget.getOriginalValue()
                self._deleteSpinBox(tomo_obj=tomo_obj)
                widget = self._createTomoObjPosition(
                    tomo_obj=tomo_obj, original_value=original_value
                )
            self._tomoObjsTableView.setIndexWidget(model_index, widget)

    # expose API
    def getEditionMode(self) -> _EditionMode:
        return self._editionOptionsWidget.getEditionMode()

    def setEditionMode(self, edition_mode: _EditionMode) -> _EditionMode:
        self._editionOptionsWidget.setEditionMode(edition_mode=edition_mode)


class EditableAxisOrderedTomoObjsModel(AxisOrderedTomoObjsModel):
    def __init__(self, axis: int, parent=None) -> None:
        super().__init__(axis, parent)
        self._headers = ["index", "tomo obj", f"axis {axis} pos (px)"]


class _TomoObjPosition(qt.QWidget):
    """
    Widget used to define tomo object position over one axis (and reset if needed)

    :param original_value: original_value of the tomo_obj along the edited axis. Used when the widget is delete and recreate
    """

    WIDGET_WIDTH = 120

    RESET_BUTTON_WIDTH = 25

    DEFAULT_VALUE_WHEN_MISSING = 0

    sigValueChanged = qt.Signal(int, int)
    """emit when the value is changed. Parameters are (old value, new value)"""

    def __init__(
        self,
        tomo_obj,
        axis_edited: int,
        parent: qt.QWidget | None = None,
        original_value: float | None = None,
    ) -> None:
        super().__init__(parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        # create spin box object
        self._spinBox = qt.QSpinBox(self)
        self._spinBox.setRange(
            numpy.iinfo(numpy.int32).min, numpy.iinfo(numpy.int32).max
        )
        self._spinBox.setSuffix("px")
        currentPos = (
            tomo_obj.stitching_metadata.get_abs_position_px(axis=axis_edited)
            or self.DEFAULT_VALUE_WHEN_MISSING
        )
        self.__originalValue = original_value or currentPos
        self.__lastValue = self.__originalValue
        self._spinBox.setValue(currentPos)
        self._spinBox.setFixedWidth(self.WIDGET_WIDTH - self.RESET_BUTTON_WIDTH)
        self.layout().addWidget(self._spinBox)

        # create option to reset value
        self._resetValueQPB = qt.QPushButton(parent=self)
        self._resetValueQPB.setFixedWidth(self.RESET_BUTTON_WIDTH)
        style = qt.QApplication.style()
        resetIcon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        self._resetValueQPB.setIcon(resetIcon)
        self.layout().addWidget(self._resetValueQPB)

        # connect signal / slot
        self._spinBox.valueChanged.connect(self._valueChanged)
        self._resetValueQPB.clicked.connect(self.resetOriginalValue)

    def _valueChanged(self, new_value):
        self.sigValueChanged.emit(
            self.__lastValue,
            new_value,
        )
        self.__lastValue = new_value

    def getOriginalValue(self) -> float:
        return self.__originalValue

    def setOriginalValue(self, value: float) -> None:
        self.__originalValue = value

    def resetOriginalValue(self) -> None:
        self._spinBox.setValue(self.getOriginalValue())
        self._positionChanged()

    def _positionChanged(self):
        self._valueChanged(new_value=self._spinBox.value())

    def setValue(self, value: float) -> None:
        self._spinBox.setValue(value)
        self.__lastValue = value

    def getValue(self) -> float:
        return self._spinBox.value()

    # expose API
    def setSingleStep(self, val: int):
        self._spinBox.setSingleStep(val)


class _EditionMode(_Enum):
    FREE = "free"
    UPSTREAM = "upstream"
    DOWNSTREAM = "downstream"

    @staticmethod
    def get_icon(cls):
        if cls == cls.UPSTREAM:
            return icons.getQIcon("edit_upstream")
        elif cls == cls.DOWNSTREAM:
            return icons.getQIcon("edit_downstream")
        elif cls == cls.FREE:
            return icons.getQIcon("free_edition")
        else:
            raise ValueError()


class _EditionOptions(qt.QWidget):

    sigReset = qt.Signal()
    sigModeChanged = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QHBoxLayout())

        self._editionModeLabel = qt.QLabel("edition mode", self)
        self.layout().addWidget(self._editionModeLabel)

        self._editionModeCB = qt.QComboBox(self)
        for mode in _EditionMode:
            self._editionModeCB.addItem(
                _EditionMode.get_icon(mode),
                mode.value,
            )
        self.layout().addWidget(self._editionModeCB)
        self._editionModeCB.setFixedHeight(30)

        self._spacer = qt.QSpacerItem(
            20, 40, qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        self.layout().addItem(self._spacer)

        style = qt.QApplication.style()
        reset_icon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        self._resetButton = qt.QPushButton(
            parent=self, text="reset all positions", icon=reset_icon
        )
        self._resetButton.setToolTip("reset all position to initial positions")
        self.layout().addWidget(self._resetButton)

        # connect signal / slot
        self._editionModeCB.currentIndexChanged.connect(self.sigModeChanged)
        self._resetButton.released.connect(self.sigReset)

    def getEditionMode(self) -> _EditionMode:
        return _EditionMode(self._editionModeCB.currentText())

    def setEditionMode(self, edition_mode: _EditionMode | str):
        self._editionModeCB.setCurrentText(_EditionMode(edition_mode).value)



# coding: utf-8

from ewokscore.task import Task as EwoksTask


class _SliceStackPlaceHolder(EwoksTask, input_names=("data",)):
    """
    Task to aggregate reconstructed slices of several scans and allow users to along them
    """

    def run(self):
        pass



from __future__ import annotations

import os
import numpy
import logging
from PIL import Image
from silx.io.url import DataUrl

from .binning import Binning
from .output import OutputFormat

from tomoscan.utils.io import filter_esrf_mounting_points
from tomoscan.esrf.scan.utils import get_data, from_bliss_original_file_to_raw

from tomwer.core.process.drac.output import DATASET_GALLERY_DIR_NAME
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase


_logger = logging.getLogger(__name__)

__all__ = [
    "DracDatasetBase",
]


class DracDatasetBase:
    """
    Abstract class for a drac dataset (that should be publish to the data portal).
    A drac dataset is defined by the following elements:
    * a directory
    * metadata
    """

    def __init__(self, data_dir: str, tomo_obj: TomwerObject) -> None:
        """
        :param data_dir: directory containing processed data
        :param tomo_obj: tomwer object referee for the processing
        """
        self._data_dir = filter_esrf_mounting_points(data_dir)
        self._tomo_obj = tomo_obj
        self._gallery_file_binning = Binning.TWO_BY_TWO
        self._gallery_output_format = OutputFormat.PNG
        self._gallery_overwrite = True

    @property
    def gallery_output_binning(self) -> Binning:
        "Binning factor to be used when saving gallery images"
        return self._gallery_file_binning

    @gallery_output_binning.setter
    def gallery_output_binning(self, binning: Binning):
        binning = Binning(binning)
        self._gallery_file_binning = binning

    @property
    def gallery_output_format(self) -> OutputFormat:
        "Output format of the images stored in the gallery"
        return self._gallery_output_format

    @gallery_output_format.setter
    def gallery_output_format(self, format: OutputFormat):
        format = OutputFormat(format)
        self._gallery_output_format = format

    @property
    def gallery_overwrite(self) -> bool:
        return self._gallery_overwrite

    @gallery_overwrite.setter
    def gallery_overwrite(self, overwrite: bool) -> None:
        self._gallery_overwrite = overwrite

    @property
    def dataset_name(self) -> str:
        """
        Return the name of the dataset to be given to the drac dataset.
        A drac dataset is store according to a (dataset_path, dataset_name). By default if several are pushed with the same key,
        only the most recent one will be stored
        """
        raise NotImplementedError("Base class")

    def get_gallery_dir(self) -> str:
        return os.path.join(self.data_dir, DATASET_GALLERY_DIR_NAME)

    def build_gallery(self):
        """callback to build the drac-dataset gallery"""
        pass

    @property
    def data_dir(self):
        return self._data_dir

    @property
    def tomo_obj(self):
        return self._tomo_obj

    @property
    def metadata(self) -> dict:
        raise NotImplementedError("Base class")

    @property
    def bliss_raw_datasets(self) -> tuple[str]:
        raise NotImplementedError("Base class")

    @staticmethod
    def from_scan_to_raws(scan: TomwerScanBase) -> tuple[str]:
        assert isinstance(
            scan, TomwerScanBase
        ), f"scan should be an instance of {TomwerScanBase}. Got {type(scan)}"
        bliss_original_files = scan.get_bliss_original_files()
        if bliss_original_files is None:
            bliss_original_files = ()
        return tuple(
            [
                from_bliss_original_file_to_raw(bliss_original_file)
                for bliss_original_file in bliss_original_files
            ]
        )

    def to_dict(self):
        return {
            "output_dir": self.data_dir,
            "drac_metadata": self.metadata,
        }

    def save_to_gallery(
        self,
        output_file_name: str,
        image: numpy.ndarray | DataUrl,
        lower_bound: float | None = None,
        upper_bound: float | None = None,
    ) -> None:
        format = self.gallery_output_format
        overwrite = self.gallery_overwrite
        binning = self.gallery_output_binning

        gallery_dir = self.get_gallery_dir()
        os.makedirs(gallery_dir, exist_ok=True)

        if isinstance(image, DataUrl):
            image = get_data(image)

        if image.ndim == 3 and image.shape[0] == 1:
            image = image.reshape(image.shape[1:])
        elif image.ndim != 2:
            raise ValueError(f"Only 2D grayscale images are handled. Got {image.shape}")

        # If both bounds are provided, apply clamping and normalization
        # If not it will use the min/max of each image
        if lower_bound is not None and upper_bound is not None:
            image = numpy.clip(image, lower_bound, upper_bound)
            image = image - lower_bound
            image *= 255.0 / (upper_bound - lower_bound)
        else:
            min_val, max_val = image.min(), image.max()
            image = image - min_val
            if max_val > min_val:  # Avoid division by zero if image is constant
                image *= 255.0 / (max_val - min_val)

        # Binning the data if required
        image = Binning._bin_data(data=image, binning=binning)

        # Convert to a PIL Image and save
        img = Image.fromarray(image.astype(numpy.uint8), mode="L")
        if not output_file_name.endswith(f".{format.value}"):
            output_file_name = output_file_name + f".{format.value}"
        if not overwrite and os.path.exists(output_file_name):
            raise OSError(f"File already exists ({output_file_name})")
        img.save(output_file_name)



# coding: utf-8
from __future__ import annotations

import datetime
import logging
from collections import OrderedDict

import h5py
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset

from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanbase import TomwerScanBase

_logger = logging.getLogger(__name__)


class LastReceivedScansDict(OrderedDict):
    """List the received scan from the first received to the last received"""

    def __init__(self, limit=None):
        """Simple structure in order to store the received last elements and
        the time of acquisition
        """
        assert limit is None or (type(limit) is int and limit > 0)
        OrderedDict.__init__(self)
        self.limit = limit

    def add(self, scan):
        assert isinstance(scan, (TomwerScanBase, BlissScan))
        self[str(scan)] = datetime.datetime.now()
        if self.limit is not None and len(self) > self.limit:
            self.pop(list(self.keys())[0])


class IgnoreProcess:
    """Simple util class to ignore a processing when using ewoks"""

    def process(self, scan):
        return scan

    __call__ = process


def get_scores(node: h5py.Group) -> dict | None:
    """read all the score from the 'results' sub group. Read the url
    and computed scores

    :return: a dictionary with (url, ComputedScore) for each node
    """
    if "results" not in node:
        _logger.warning(f"no results found in {node}")
        return None
    else:
        res_node = node["results"]
        scores = {}
        for res_item in res_node:
            score_node = res_node[res_item]
            try:
                float(res_item)
            except Exception:
                is_a_score = False
            else:
                is_a_score = True

            if not is_a_score:
                continue

            try:
                url = DataUrl(
                    file_path=node.file.filename,
                    data_path="/".join((score_node.name, "reconstructed_slice")),
                    scheme="silx",
                )
                score = ComputedScore(
                    tv=h5py_read_dataset(score_node["total variation"]),
                    std=h5py_read_dataset(score_node["standard deviation"]),
                )
                scores[float(res_item)] = (url, score)
            except Exception as e:
                _logger.warning(f"Unable to load {score_node.name}. Reason is {e}")
        scores = dict(sorted(scores.items(), key=lambda item: item[0]))

        return scores



from silx.utils.deprecation import deprecated

from tomwer.core.futureobject import TomwerScanBase


@deprecated(replacement="TomwerScanBase", since_version="1.0")
class FutureTomwerScan(TomwerScanBase):
    pass



# coding: utf-8
from __future__ import annotations

import os
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt
from tomoscan.series import Series

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import MockEDF, MockNXtomo
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.gui.control.series.seriescreator import (
    SeriesDefinition,
    SeriesHistoryDialog,
    SeriesManualControlDialog,
    SeriesManualFromTomoObj,
    SeriesTree,
    SeriesWidget,
)


class _MockScanBase:
    def init(self):
        n_hdf5_scan = 11
        self._scans = []
        self._root_dir = tempfile.mkdtemp()
        for i_scan in range(n_hdf5_scan):
            scan_path = os.path.join(self._root_dir, f"scan_{i_scan}")
            scan = MockNXtomo(scan_path=scan_path, n_proj=10, n_ini_proj=0).scan
            self._scans.append(scan)
        n_edf_scan = 4
        for i_scan in range(n_edf_scan):
            scan_path = os.path.join(self._root_dir, f"scan_{i_scan}")
            scan = MockEDF.mockScan(
                scanID=scan_path,
                nRadio=10,
                dim=10,
            )
            self._scans.append(scan)

    def close(self):
        shutil.rmtree(self._root_dir)


class _MockSeriesBase(_MockScanBase):
    def init(self):
        super().init()
        self._seriesList = [
            Series("series1", self._scans[0:1], use_identifiers=True),
            Series("series2", self._scans[1:5], use_identifiers=False),
            Series("series3", self._scans[5:11], use_identifiers=True),
            Series("series4", self._scans[8:11], use_identifiers=True),
            Series("series5", self._scans[-2:-1], use_identifiers=False),
            Series("series6", self._scans[-5:], use_identifiers=True),
        ]

    def close(self):
        self._seriesList.clear()
        super().close()


class TestSeriesTree(TestCaseQt, _MockSeriesBase):
    """Test the SeriesTree widget"""

    def setUp(self):
        super().setUp()
        super().init()
        self._widget = SeriesTree()

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        super().close()
        super().tearDown()

    def test_add_remove(self):
        """Test adding and removing series"""
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)
        for series in self._seriesList:
            self._widget.addSeries(series)
        assert self._widget.n_series == 6
        self._widget.removeSeries(self._seriesList[3])
        assert self._widget.n_series == 5
        self._widget.removeSeries(self._seriesList[2])
        self._widget.removeSeries(self._seriesList[1])
        self._widget.removeSeries(self._seriesList[0])
        # make sure no error is raised if we try to remove twine the same serie
        self._widget.removeSeries(self._seriesList[0])
        assert self._widget.n_series == 2
        self._widget.addSeries(self._seriesList[1])
        self._widget.addSeries(self._seriesList[2])
        assert self._widget.n_series == 4

    def test_selection(self):
        """Test selection of the SerieTree"""
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)
        for series in self._seriesList:
            self._widget.addSeries(series)

        selection = (self._seriesList[2], self._seriesList[3])
        self._widget.setSelectedSeries(selection)
        assert self._widget.getSelectedSeries() == selection
        self._widget.clearSelection()
        assert self._widget.getSelectedSeries() == ()


class TestSeriesHistoryDialog(TestCaseQt, _MockSeriesBase):
    """Test the SeriesHistoryDialog"""

    def setUp(self):
        super().setUp()
        super().init()
        self._widget = SeriesHistoryDialog()

        # create listener for the nabu widget
        self.signal_listener = SignalListener()

        # connect signal / slot
        self._widget.sigSeriesSend.connect(self.signal_listener)

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        super().close()
        super().tearDown()

    def test(self):
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)
        for series in self._seriesList:
            self._widget.addSeries(series)

        selection = (self._seriesList[0], self._seriesList[4])
        self._widget.setSelectedSeries(selection)
        assert self._widget.getSelectedSeries() == selection
        assert self.signal_listener.callCount() == 0
        self._widget._sendButton.clicked.emit()
        self.qapp.processEvents()
        assert self.signal_listener.callCount() == 2
        assert self._widget.getSelectedSeries() == selection
        self.signal_listener.clear()
        self._widget._clearButton.clicked.emit()
        assert self._widget.getSelectedSeries() == ()
        assert self.signal_listener.callCount() == 0
        self._widget._sendButton.clicked.emit()
        assert self.signal_listener.callCount() == 0


class TestSeriesDefinition(TestCaseQt):
    def setUp(self):
        super().setUp()
        self._widget = SeriesDefinition()

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        super().tearDown()

    def test_manual_selection(self):
        self._widget.setMode("manual")
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)
        assert not self._widget._automaticDefWidget.isVisible()
        assert self._widget._manualDefWidget.isVisible()

    def test_automatic_selection(self):
        self._widget.setMode("auto")
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)
        assert self._widget._automaticDefWidget.isVisible()
        assert not self._widget._manualDefWidget.isVisible()


class TestSeriesManualDefinitionDialog(TestCaseQt, _MockScanBase):
    """Test interaction with the series manual definition"""

    def setUp(self):
        self._widget = SeriesManualControlDialog()
        super().setUp()
        super().init()

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        super().close()
        super().tearDown()

    def test(self):
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)

        self._widget.setSeriesName("series test")
        self.qapp.processEvents()
        assert self._widget.getSeriesName() == "series test"

        self._widget._mainWidget._newSeriesWidget._serieTree.rootItem.setText(
            0, "new series"
        )
        self.qapp.processEvents()
        assert self._widget.getSeriesName() == "new series"

        for scan in self._scans[:5]:
            self._widget.addScan(scan)

        self.assertEqual(self._widget.n_tomo_objs, len(self._scans[:5]))
        self._widget.removeScan(self._scans[0])
        self.assertEqual(self._widget.n_tomo_objs, len(self._scans[:5]) - 1)

        series_scan = tuple(self._scans[1:5])
        assert isinstance(series_scan, tuple)

        current_series = self._widget.getSeries(use_identifiers=True)
        assert isinstance(current_series, Series)
        series_test_1 = Series(
            name="new series", iterable=series_scan, use_identifiers=True
        )

        self.assertEqual(series_test_1, current_series)
        series_test_2 = Series(name="test", iterable=series_scan, use_identifiers=True)
        assert series_test_2.name == "test"
        assert current_series.name == "new series"
        self.assertNotEqual(series_test_2, current_series)

        self._widget.setSelectedScans([self._scans[2]])
        self._widget.getSelectedScans() == (self._scans[2],)
        self._widget.removeSelectedScans()
        self.assertEqual(self._widget.n_tomo_objs, len(self._scans[:5]) - 2)
        self._widget.getSelectedScans() == tuple()

        self._widget.clearSeries()
        self.assertEqual(self._widget.n_tomo_objs, 0)
        self.assertEqual(Series(name="new series"), self._widget.getSeries())

        # test adding an nx file
        hdf5_scan = self._scans[0]
        assert isinstance(hdf5_scan, NXtomoScan)
        self._widget.addScanFromNxFile(hdf5_scan.master_file)
        self.assertEqual(self._widget.n_tomo_objs, 1)


class TestSeriesWidget(TestCaseQt, _MockSeriesBase):
    """
    Test the SeriesWidget
    """

    def setUp(self):
        super().setUp()
        super().init()
        self._widget = SeriesWidget()
        # create listeners
        self.signal_send_series_listener = SignalListener()
        self.signal_serie_changed_listener = SignalListener()
        self.signal_history_changed_listener = SignalListener()

        # connect signal / slot
        self._widget.sigSeriesSend.connect(self.signal_send_series_listener)
        self._widget.sigCurrentSeriesChanged.connect(self.signal_serie_changed_listener)
        self._widget.sigHistoryChanged.connect(self.signal_history_changed_listener)

    def tearDown(self):
        self._widget.sigSeriesSend.disconnect(self.signal_send_series_listener)
        self._widget.sigCurrentSeriesChanged.disconnect(
            self.signal_serie_changed_listener
        )
        self._widget.sigHistoryChanged.disconnect(self.signal_history_changed_listener)

        self.signal_send_series_listener = None
        self.signal_serie_changed_listener = None
        self.signal_history_changed_listener = None

        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        super().close()
        super().tearDown()

    def test(self):
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)
        self._seriesList[3].name = "toto series"
        for series in self._seriesList[2:5]:
            self._widget.getHistoryWidget().addSeries(series)
        self._widget.setMode("history")
        self._widget.setMode("series definition", "manual")
        self._widget.getDefinitionWidget().getManualDefinitionWidget().setSeriesName(
            "new series"
        )

        self.assertEqual(
            len(
                self._widget.getDefinitionWidget()
                .getManualDefinitionWidget()
                .getSeries()
            ),
            0,
        )

        self._widget.getHistoryWidget().setSelectedSeries(
            [
                self._seriesList[3],
            ]
        )
        assert len(self._widget.getHistoryWidget().getSelectedSeries()) == 1

        self.signal_serie_changed_listener.clear()
        self._widget.getHistoryWidget().editSelected()
        self.qapp.processEvents()
        assert self.signal_serie_changed_listener.callCount() == 1

        self.assertEqual(
            self._widget.getDefinitionWidget()
            .getManualDefinitionWidget()
            .getSeries()
            .name,
            "toto series",
        )

        self.assertEqual(
            self._widget.getDefinitionWidget()
            .getManualDefinitionWidget()
            .getSeries(use_identifiers=True),
            self._seriesList[3],
        )

        self.signal_serie_changed_listener.clear()
        self._widget.getDefinitionWidget().getManualDefinitionWidget().addToCurrentSeries(
            self._scans[0]
        )
        self.qapp.processEvents()
        assert self.signal_serie_changed_listener.callCount() == 1

        expected_scans = self._scans[8:11]
        expected_scans.append(self._scans[0])
        expected_series = Series("toto series", expected_scans, use_identifiers=True)

        self.assertEqual(
            self._widget.getDefinitionWidget()
            .getManualDefinitionWidget()
            .getSeries(use_identifiers=True),
            expected_series,
        )

        # check send edited serie
        self.signal_history_changed_listener.clear()
        self.signal_send_series_listener.clear()
        self.qapp.processEvents()

        # check send selected from the history
        self.signal_send_series_listener.clear()
        self._widget.getHistoryWidget().setSelectedSeries(
            [
                self._seriesList[4],
            ]
        )
        assert len(self._widget.getHistoryWidget().getSelectedSeries()) == 1
        self._widget.getHistoryWidget().sendSelected()
        assert self.signal_send_series_listener.callCount() == 1


class TestSerieManualFromTomoObj(TestCaseQt):
    """
    test the SerieManualFromTomoObj widget
    """

    def setUp(self):
        super().setUp()
        self._tmp_dir = tempfile.mkdtemp()
        self._widget = SeriesManualFromTomoObj()
        self._volume_1 = HDF5Volume(
            file_path=os.path.join(self._tmp_dir, "vol1.hdf5"),
            data_path="data",
        )
        self._volume_2 = HDF5Volume(
            file_path=os.path.join(self._tmp_dir, "vol2.hdf"),
            data_path="data",
        )
        self._volume_3 = HDF5Volume(
            file_path=os.path.join(self._tmp_dir, "vol3.nx"),
            data_path="data",
        )
        self._scan_1 = MockNXtomo(
            scan_path=os.path.join(self._tmp_dir, "scan_1"), n_proj=10, n_ini_proj=10
        ).scan
        self._scan_2 = MockNXtomo(
            scan_path=os.path.join(self._tmp_dir, "scan_2"), n_proj=10, n_ini_proj=10
        ).scan

    def tearDown(self):
        shutil.rmtree(self._tmp_dir)
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        return super().tearDown()

    def test(self):
        self._widget.show()
        for tomo_obj in (
            self._volume_1,
            self._volume_2,
            self._volume_3,
            self._scan_1,
            self._scan_2,
        ):
            self._widget.addTomoObj(tomo_obj)

        current_serie = self._widget.getSeries()
        assert isinstance(current_serie, Series)
        assert len(current_serie) == 0

        for tomo_obj in (self._volume_1, self._volume_2):
            self._widget.addToCurrentSeries(tomo_obj)

        current_serie = self._widget.getSeries()
        assert isinstance(current_serie, Series)
        assert len(current_serie) == 2
